
import os
import subprocess
import shutil
from timeit import default_timer as timer

import logging
log = logging.getLogger(__name__)

from partis.schema import (
  Loc,
  required,
  optional,
  derived,
  is_sequence,
  is_mapping,
  is_evaluated,
  is_valued,
  is_valued_type,
  is_optional,
  PyEvaluated,
  CheetahEvaluated,
  BoolPrim,
  IntPrim,
  FloatPrim,
  StrPrim,
  SeqPrim,
  MapPrim,
  UnionPrim,
  StructValued,
  schema_declared,
  EvaluatedContext )

from partis.utils import (
  ModelHint,
  Loc,
  HINT_LEVELS_DESC,
  indent_lines )

from partis.utils.hint import (
  DATA_FORMATS )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
hint_declared = schema_declared( tag = 'hint' )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
HintList = SeqPrim(
  doc = "List of hints",
  item = hint_declared,
  default_val = list() )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class HintLoc( StructValued ):
  """A schema for serializing ModelHints
  """

  schema = dict(
    tag = 'loc',
    default_val = derived,
    struct_proxy = 'filename' )

  filename = StrPrim(
    doc = "Referenced location",
    default_val = "",
    max_lines = 1 )

  line = IntPrim( default_val = 0 )
  col = IntPrim( default_val = 0 )

  path = SeqPrim(
    default_val = list(),
    item = UnionPrim(
      cases = [
        IntPrim(),
        StrPrim(
          max_lines = 1 )] ) )

  owner = StrPrim(
    doc = "Owner of reference",
    default_val = "",
    max_lines = 1 )

  time = FloatPrim( default_val = 0 )

  #-----------------------------------------------------------------------------
  def _cast( self ):
    """Converts instance of this to instance of a regular ModelHint

    Returns
    :class:`ModelHint <partis.utils.hint.ModelHint>`
    """
    d = self._schema.encode(
      val = self,
      loc = self._loc,
      no_defaults = True )

    return Loc.from_dict(d)

  #-----------------------------------------------------------------------------
  def __rich__( self ):
    return self._cast().fmt( with_rich = True )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class Hint( StructValued ):
  """A schema for serializing ModelHints
  """

  schema = dict(
    declared = hint_declared,
    default_val = optional,
    struct_proxy = 'msg' )

  level = StrPrim(
    char_case = 'upper',
    default_val = "INFO",
    restricted = [ k.upper() for k,v in HINT_LEVELS_DESC.items() ],
    doc = "\n".join([ f"- ``'{k.upper()}'``: {indent_lines(2, v, start = 1)}" for k,v in HINT_LEVELS_DESC.items() ]) )

  msg = StrPrim(
    doc = "Message",
    default_val = "" )

  data = StrPrim(
    doc = "Data",
    default_val = "" )

  format = StrPrim(
    doc = "Data format",
    default_val = DATA_FORMATS[0],
    restricted = DATA_FORMATS )


  loc = HintLoc

  hints = HintList

  #-----------------------------------------------------------------------------
  def model_hint( self ):
    """Converts instance of this to instance of a regular ModelHint

    Returns
    :class:`ModelHint <partis.utils.hint.ModelHint>`
    """
    d = self._schema.encode(
      val = self,
      loc = self._loc,
      no_defaults = True )

    if d is None:
      return d

    return ModelHint.from_dict(d)




from copy import copy

import re
import uuid
import sys

from abc import ABCMeta
import inspect

import logging
log = logging.getLogger(__name__)

from partis.pyproj import (
  hash_sha256 )

from pygments.lexers import get_lexer_by_name

from partis.utils import (
  cached_property,
  adict,
  odict,
  rdict,
  rlist,
  StringFunction,
  mapping_attrs,
  protected_attr,
  indent_lines,
  fmt_base_or_type,
  fmt_class_name,
  _fmt_class_name,
  fmt_attr_doc,
  make_dynamic_class_name )


from .base import (
  SchemaError,
  SchemaNameError,
  SchemaDeclaredError,
  SchemaDefinitionError,
  SchemaValidationError,
  SchemaHint,
  Loc,
  assert_valid_name,
  assert_valid_path,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_optional,
  is_required,
  is_derived,
  is_similar_value_type,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  is_provider,
  is_evaluated_class,
  is_evaluated,
  is_valued,
  is_valued_type,
  any_schema )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class ProviderMeta( ABCMeta ):
  """Meta class for evaluation provider
  """
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class ProviderSupport:
  """Information about a single supported evaluation by a provider

  Parameters
  ----------
  name : str
    A unique name of supported evaluation
  lexer : NoneType | str | Lexer
    Pygments lexer that may be used to apply syntax highlighting.
    If None, no highlighting will be applied.
  doc : NoneType | str
    Extra documentation about the support
  """

  #-----------------------------------------------------------------------------
  def __init__( self,
    name,
    lexer = None,
    doc = None ):

    if isinstance( lexer, str ):
      lexer = get_lexer_by_name( lexer )

    if doc is None:
      doc = ""

    self._name = str(name)
    self._lexer = lexer
    self._doc = inspect.cleandoc( doc )

  #-----------------------------------------------------------------------------
  @property
  def name( self ):
    return self._name

  #-----------------------------------------------------------------------------
  @property
  def lexer( self ):
    return self._lexer

  #-----------------------------------------------------------------------------
  @property
  def doc( self ):
    return self._doc

  #-----------------------------------------------------------------------------
  def __str__( self ):
    return self._name

  #-----------------------------------------------------------------------------
  def __lt__( self, other ):
    return self._name < str(other)

  #-----------------------------------------------------------------------------
  def __eq__( self, other ):
    return self._name == str(other)

  #-----------------------------------------------------------------------------
  def __hash__( self ):
    return hash(self._name)

  #-----------------------------------------------------------------------------
  @cached_property
  def schema_hash_node( self ):
    """urlsafe Base64 encoded hash
    """

    return hash_sha256( self._name.encode('utf-8') )[0]

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class Provider( metaclass = ProviderMeta ):

  #-----------------------------------------------------------------------------
  def __init__( self ):
    pass

  #-----------------------------------------------------------------------------
  @property
  def supported( self ):
    """Supported types of evaluations

    Returns
    -------
    : dict[ str, :class:`ProviderSupport`]
    """
    raise NotImplementedError("Derived provider must implement `supported`")

  #-----------------------------------------------------------------------------
  def check( self, src ):
    """Checks whether a particular source is supported

    Parameters
    ----------
    src : object
      The source data to be checked for supporting the evaluation

    Returns
    -------
    : None | tuple[ :class:`ProviderSupport`, object ]
      The evaluation support for the source data, and a (possibly) modified source
      object such as removal of any escape sequence.
      If None, evaluation is not supported
    """
    raise NotImplementedError("Derived provider must implement `check`")

  #-----------------------------------------------------------------------------
  def escaped( self, support, src ):
    """Modifies source data to apply escaping
    """

    raise NotImplementedError("Derived provider must implement `escaped`")

  #-----------------------------------------------------------------------------
  def lint( self,
    schema,
    src,
    loc = None,
    locals = None,
    module = None,
    logger = None ):
    """Lints source data

    Parameters
    ----------
    schema : :class:`Schema <partis.schema.Schema>`
      Schema for evaluation result
    src : object
      The source data to be evaluated
    loc : None | Loc
    locals : dict[ str, object ] | None
      Defines local variables to use for evaluation.
    module : None | str | ModuleType
      Effective python module for evaluation context, such as performing relative imports
    logger : logging.Logger
      A logger object to direct expression logs/prints instead of stdout

    Returns
    -------
    errors : list[SchemaHint]
      List of any errors to report from linting of source
    """
    return list()

  #-----------------------------------------------------------------------------
  def eval( self,
    schema,
    src,
    loc = None,
    locals = None,
    module = None,
    logger = None ):
    """Evaluates source data

    Parameters
    ----------
    schema : :class:`Schema <partis.schema.Schema>`
      Schema for evaluation result
    src : object
      The source data to be evaluated
    loc : None | Loc
    locals : dict[ str, object ] | None
      Defines local variables to use for evaluation.
    module : None | str | ModuleType
      Effective python module for evaluation context, such as performing relative imports
    logger : logging.Logger
      A logger object to direct expression logs/prints instead of stdout

    Returns
    -------
    value : object
      Value resulting from evaluation.
      If there are no un-evaluated values, then will return a shallow copy of
      the current data.

    """
    raise NotImplementedError("Derived provider must implement `eval`")

  #-----------------------------------------------------------------------------
  @cached_property
  def schema_hash_node( self ):
    """urlsafe Base64 encoded hash
    """

    return hash_sha256( '\n'.join([
      f'{k}, {v.schema_hash_node}'
      for k,v in self.supported.items()  ]).encode('utf-8') )[0]


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class ProviderUnion( Provider ):

  #-----------------------------------------------------------------------------
  def __init__( self, providers ):
    super().__init__()

    support_map = dict()
    supported = dict()

    for i, p in enumerate(providers):
      if not isinstance(p, Provider):
        raise SchemaDefinitionError(
          f"provider {i} must be instance of Provider: {p}")

      for k, s in p.supported.items():
        if k in supported:
          if supported[k] != s:
            raise SchemaDefinitionError(
              f"provider support '{k}' redefined: {s}")
        else:
          supported[k] = s
          support_map[s] = p

    self._p_providers = providers
    self._p_supported = supported
    self._p_support_map = support_map

  #-----------------------------------------------------------------------------
  @property
  def supported( self ):
    return self._p_supported

  #-----------------------------------------------------------------------------
  def check( self, src ):
    for i, p in enumerate(self._p_providers):
      res = p.check(src)

      if res:
        return res

    return None

  #-----------------------------------------------------------------------------
  def escaped( self, support, src ):
    if support in self._p_support_map:
      return self._p_support_map[support].escaped( support, src )

    raise ValueError(f"`support` must be one of {self.supported.values()}: {support}")

  #-----------------------------------------------------------------------------
  def lint( self,
    schema,
    src,
    loc = None,
    locals = None,
    module = None,
    logger = None ):

    supported = self.check( src )

    if not supported:
      raise SchemaEvaluationError(f"No evaluation support: {src}")

    support, _src = supported

    return self._p_support_map[support].lint(
      schema = schema,
      src = src,
      loc = loc,
      locals = locals,
      module = module,
      logger = logger )

  #-----------------------------------------------------------------------------
  def eval( self,
    schema,
    src,
    loc = None,
    locals = None,
    module = None,
    logger = None ):

    supported = self.check( src )

    if not supported:
      raise SchemaEvaluationError(f"No evaluation support: {src}")

    support, _src = supported

    return self._p_support_map[support].eval(
      schema = schema,
      src = src,
      loc = loc,
      locals = locals,
      module = module,
      logger = logger )

  #-----------------------------------------------------------------------------
  @cached_property
  def schema_hash_node( self ):

    return hash_sha256( '\n'.join([
      f'{k}, {v.schema_hash_node}'
      for k,v in self.supported.items()  ]).encode('utf-8') )[0]

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class EvaluatedContextMeta( ABCMeta ):
  """Defines a context for evaluating expressions.

  Parameters
  ----------
  id : string
    Identifier for this context
  doc : NoneType | string
  """

  #-----------------------------------------------------------------------------
  def __new__(mcls,
    name,
    bases,
    namespace,
    id = None,
    doc = None ):

    assert_valid_name( name )

    base_id = set()

    for base in bases:

      if isinstance( base, EvaluatedContextMeta ):
        base_id.add( base.id )

    if id is None and len(base_id) > 0:

      if len(base_id) > 1:
        raise SchemaDefinitionError(
          f"base class `id` is ambiguous: {base_id}")

      id = next(iter(base_id))

    # NOTE: this is intended behaviour even when id is None, resulting in
    # id -> 'None', denoting a general category with little restrictions
    id = str(id)

    # format docstring
    if doc is None:
      doc = ""
    else:
      doc = inspect.cleandoc( doc )

    if '__doc__' in namespace and namespace['__doc__']:
      _doc = inspect.cleandoc( str(namespace['__doc__']) )

      if doc:
        doc += '\n\n' + _doc
      else:
        doc = _doc


    namespace = { **namespace, **dict(
      _p_doc = doc,
      __doc__ = StringFunction(),
      _p_id = id ) }

    cls = super().__new__( mcls, name, bases, namespace )

    cls.__doc__._func = cls._get_doc

    return cls

  #-----------------------------------------------------------------------------
  def __init__( cls,
    name,
    bases,
    namespace,
    id = None,
    doc = None ):

    super().__init__( name, bases, namespace )

  #-----------------------------------------------------------------------------
  @classmethod
  def __prepare__( mcls, name, bases, *args, **kwargs ):
    return odict()

  #-----------------------------------------------------------------------------
  def _get_doc( cls ):
    return cls.doc

  #-----------------------------------------------------------------------------
  @property
  def doc( cls ):
    return cls._p_doc

  #-----------------------------------------------------------------------------
  @property
  def id( cls ):
    return cls._p_id

  #-----------------------------------------------------------------------------
  def __str__( cls ):
    return cls._p_id

  #-----------------------------------------------------------------------------
  def __hash__( cls ):
    return hash(cls._p_id)

  #-----------------------------------------------------------------------------
  def __eq__( cls, other ):

    return cls._p_id == str(other)

  #-----------------------------------------------------------------------------
  def __ne__( cls, other ):

    return cls._p_id != str(other)

  #-----------------------------------------------------------------------------
  @cached_property
  def schema_hash_node( cls ):
    """urlsafe Base64 encoded hash
    """

    return hash_sha256(cls._p_id.encode('utf-8'))[0]

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class EvaluatedContext(
  metaclass = EvaluatedContextMeta ):
  """Defines a context for evaluating expressions.

  Parameters
  ----------
  module : None | str | ModuleType
    Module relative to which the evaluation is to occur
  """

  #-----------------------------------------------------------------------------
  def __init__(self, *, module = None ):

    self._p_module = module

  #-----------------------------------------------------------------------------
  @property
  def module( self ):
    return self._p_module

  #-----------------------------------------------------------------------------
  def locals( self, *,
    schema ):
    """Provides the local variables to use when evaluating

    Parameters
    ----------
    schema : :class:`SchemaPrim <partis.schema.prim.base.SchemaPrim>` | :class:`SchemaStruct <partis.schema.struct.SchemaStruct>`
      Schema for the resulting value.

    Returns
    -------
    :class:`Mapping`
    """

    return dict()

  #-----------------------------------------------------------------------------
  def __call__( self, *,
    schema,
    parent,
    key ):
    """Generates a new context for a child of the current valued object.

    .. note::
      New context instance must be of the same class as the current context

    Parameters
    ----------
    schema : :class:`SchemaPrim <partis.schema.prim.base.SchemaPrim>` | :class:`SchemaStruct <partis.schema.struct.SchemaStruct>`
      Schema for the parent value.
    parent : :class:`Valued <partis.schema.valued.Valued>` | :class:`SchemaStruct <partis.schema.struct.SchemaStruct>`
      Parent valued object to which the current :class:`EvaluatedNamespace` was
      originally generated.

      .. note::

        The parent value will only contain those values that have been
        evaluated up to the point at which the current value is being evaluated.
        If those previous evaluations have a different context, they will
        may remain un-evaluated.

    key : str | int
      Key (str) of mapping, or index (int) of sequence, within the ``parent``
      that the returned ``EvaluatedContext`` should correspond.

    Returns
    -------
    :class:`EvaluatedContext`
    """

    return self

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class EvaluatedMeta( ABCMeta ):
  """Meta-class for a variable type that must be evaluated

  Parameters
  ----------
  name : str
    Class name of type
  bases : list[type]
    Base classes of type
  namespace : dict[str, object]
  provider : NoneType | :class:`Provider <partis.schema.eval.Provider>`
    Provider for evaluating expressions.
  context : NoneType | :class:`EvaluatedContextMeta`
    Sets the context for which these expressions is to be evaluated.
    A value of None will cause the expression to be evaluated upon any call.
    Otherwise, the expression is evaluated only when called with the specified
    context.
  """
  #-----------------------------------------------------------------------------
  def __new__( mcls,
    name,
    bases,
    namespace,
    provider = None,
    context = None ):

    assert_valid_name( name )

    base_provider = list()
    base_context = list()

    for base in bases:

      if is_evaluated_class( base ):
        if base.provider and base.provider not in base_provider:
          base_provider.append( base.provider )

        if base.context and base.context not in base_context:
          base_context.append( base.context )

    if provider is None and len(base_provider) > 0:

      if len(base_provider) > 1:
        provider = ProviderUnion(base_provider)
      else:
        provider = next(iter(base_provider))

    if provider is not None and not is_provider( provider ):
        raise SchemaDefinitionError(
          f"`provider` must be instance of `Provider`: {provider}")

    if context is None:
      if len(base_context) > 0:

        if len(base_context) > 1:
          raise SchemaDefinitionError(
            f"base class `context` is ambiguous: {base_context}")

        context = next(iter(base_context))

      else:
        context = EvaluatedContext

    if not isinstance( context, EvaluatedContextMeta ):
      raise SchemaDefinitionError(
        f"`context` must be an `EvaluatedContextMeta`: {context}")


    if '__doc__' not in namespace:
      lines = list()

      if provider is None:
        supported = "Evaluation not supported."
      else:
        supported = "\n".join([
          'Supported evaluations:\n',
          *[f"* {s}" for s in provider.supported.values()] ])

      lines.extend([
        fmt_attr_doc(
          name = 'context',
          prefix = 'property',
          typename = _fmt_class_name(context),
          val = context,
          doc = context.doc ),
        fmt_attr_doc(
          name = 'provider',
          prefix = 'property',
          typename = type(provider),
          val = ...,
          doc = supported ) ])


      namespace['__doc__'] = "\n".join(lines)

    namespace = odict( **namespace,
      _p_num_subclass = 0,
      _p_provider = provider,
      _p_context = context )

    cls = super().__new__( mcls, name, bases, namespace )

    return cls

  #-----------------------------------------------------------------------------
  def __init__( cls,
    name,
    bases,
    namespace,
    provider = None,
    context = None ):

    super().__init__( name, bases, namespace )

  #-----------------------------------------------------------------------------
  def __or__( cls, other ):

    module, name = make_dynamic_class_name(
      default_name = cls.__name__ + '_' + other.__name__ )

    return cls.subclass(
      name = name,
      module = module,
      bases = (other,) )

  #-----------------------------------------------------------------------------
  @property
  def provider( cls ):
    """:class:`Provider <partis.schema.eval.Provider>` : Provider for evaluating expressions.
    """
    return cls._p_provider

  #-----------------------------------------------------------------------------
  @property
  def context( cls ):
    """NoneType | :class:`EvaluatedContextMeta`
    """
    return cls._p_context

  #-----------------------------------------------------------------------------
  @property
  def supported( cls ):
    """Supported types of evaluations

    Returns
    -------
    : dict[ str, :class:`ProviderSupport`]
    """

    if cls.provider is None:
      return dict()

    return cls.provider.supported

  #-----------------------------------------------------------------------------
  def check( cls, src ):
    """Checks whether a particular source is supported

    Parameters
    ----------
    src : object
      The source data to be checked

    Returns
    -------
    : NoneType | tuple[ :class:`ProviderSupport`, object ]
      The evaluation support for the source data, and a (possibly) modified source
      object such as removal of any escape sequence.
      If None, evaluation is not supported
    """
    if cls.provider is None:
      return None

    return cls.provider.check( src )

  #-----------------------------------------------------------------------------
  def escaped( cls, support, src ):
    """Modifies source data to apply escaping
    """

    if cls.provider is None:
      return src

    return cls.provider.escaped( support, src )

  #-----------------------------------------------------------------------------
  def subclass( cls,
    name = None,
    module = None,
    module_set = None,
    bases = None,
    **kwargs ):
    """
    Parameters
    ----------
    name : NoneType | str
      Name of subclass.
      If None, will generate a name automatically.
    module : None | str | ModuleType
      Module to use for the subclass.
      If None, the module will be determined from the caller frame.
      If a string, the module must be importable.
    module_set : None | bool
      If True (or None), adds the subclass to the module `module`.
      If False, the return value will be the only reference to the subclass.
    **kwargs :
      Keyword arguments passed
    """

    module, name = make_dynamic_class_name(
      default_name = cls.__name__,
      module = module,
      name = name )

    cls._p_num_subclass += 1

    if bases is None:
      bases = tuple()

    bases = ( cls, ) + bases

    namespace = type(cls).__prepare__(
      name,
      bases,
      **kwargs )

    namespace['__module__'] = module.__name__

    sub_cls = type(cls)(
      name,
      bases,
      namespace,
      **kwargs )

    if module_set is None or module_set:

      if hasattr( module, sub_cls.__name__ ):
        warnings.warn(f"Overriding module attribute with class definition: {sub_cls.__name__}")

      setattr( module, sub_cls.__name__, sub_cls )



    return sub_cls

  #-----------------------------------------------------------------------------
  def _schema_hash_node( cls ):

    lines = list()

    for attr in [
      'provider',
      'context' ]:

      val = getattr( cls, attr )
      hash = None

      if hasattr( val, 'schema_hash_node' ):
        hash = val.schema_hash_node
      else:
        hash = hash_sha256( str(val).encode('utf-8') )[0]

      lines.append( f'{hash}' )

    return lines

  #-----------------------------------------------------------------------------
  @cached_property
  def schema_hash_node( cls ):
    """urlsafe Base64 encoded hash
    """

    return hash_sha256(
      '\n'.join( cls._schema_hash_node() ).encode('utf-8') )[0]



import logging
log = logging.getLogger( __name__ )

from copy import copy

from partis.utils import (
  adict,
  adict_struct,
  odict,
  attrs_modify )


from partis.utils.special import (
  required,
  optional )

from partis.schema_meta.valued import (
  ValuedMeta )

from partis.schema_meta.base import (
  SchemaError,
  SchemaHint,
  Loc,
  assert_valid_name,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_special,
  is_optional,
  is_required,
  is_derived,
  is_similar_value_type,
  is_schema_prim,
  is_schema_declared,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  is_evaluated_class,
  is_evaluated,
  is_valued,
  is_valued_type,
  any_schema )

from partis.schema import (
  EvaluatedContext )

from .valued import (
  Valued,
  get_schema_loc,
  get_init_val,
  get_src_val )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SeqValued( Valued, list ):
  """Container for fully or partially evaluated sequence (list, array) data

  Parameters
  ----------
  val : list[object]
    Source data to be evaluated
  schema : :class:`Schema <partis.schema.struct.SchemaStruct>` | :class:`SchemaPrim <partis.schema.prim.base.SchemaPrim>`
  loc : :class:`Loc <partis.schema_meta.base.Loc>`
    Location information of source data (E.G. file, line/column number)
  """

  #-----------------------------------------------------------------------------
  def __new__( cls,
    val = None,
    schema = None,
    loc = None,
    bias = None ):

    self = list.__new__( cls )

    return self

  #-----------------------------------------------------------------------------
  def __init__( self,
    val = None,
    schema = None,
    loc = None,
    bias = None ):

    _val = get_init_val(
      val = val,
      schema = schema,
      default_val = list() )

    list.__init__( self, _val )

    Valued.__init__( self,
      val = val,
      schema = schema,
      loc = loc,
      bias = bias )

  #-----------------------------------------------------------------------------
  def __setitem__( self, key, val ):
    if is_evaluated( self._src ):
      # NOTE: If source is "evaluated", then encoding will not preserve set values
      raise SchemaError(f"Cannot set items on un-evaluated object: {key}")

    loc = None

    if is_valued_type( val ):
      loc = val._loc

    if self._schema is not None:
      _schema = self._schema.item.schema

      if not any_schema(val, [_schema]):
        val = self._schema.item.schema.decode(
          val = val,
          loc = loc )

    super().__setitem__( key, val )

  #-----------------------------------------------------------------------------
  def append( self, val ):

    if self._schema is not None:
      val = self._schema.item.schema.decode( val )

    super().append( val )

  #-----------------------------------------------------------------------------
  def extend( self, vals ):

    if self._schema is not None:
      _schema = self._schema.item.schema

      vals = [
        val if any_schema(val, [_schema]) else self._schema.item.schema.decode( val )
        for val in vals ]

    super().extend( vals )

  #-----------------------------------------------------------------------------
  def __str__( self ):
    if is_evaluated( self._src ):
      return str(self._src)

    return f"{type(self).__name__}({list(self)}"

  #-----------------------------------------------------------------------------
  def __repr__( self ):
    return str(self)

  #-----------------------------------------------------------------------------
  @property
  def _encode( self ):

    if is_evaluated( self._src ):
      return self._src._encode

    if self._schema is not None:
      return self._schema.encode( list(self), self._loc )

    return [ v._encode if ( is_valued_type(v) or is_evaluated(v) ) else v for v in self ]



from partis.schema_meta.base import (
  Loc )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def as_load( loads ):

  def load(
    file,
    *args,
    **kwargs ):

    if 'loc' not in kwargs:
      kwargs['loc'] = Loc(
        filename = file )

    with open( file, 'rb' ) as fp:
      src = fp.read()

    src = src.decode( 'utf-8', errors = 'replace' )

    return loads( src, *args, **kwargs )

  return load

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def as_dump( dumps ):

  def dump(
    file,
    *args,
    **kwargs ):

    src = dumps(
      *args,
      **kwargs )

    src = src.encode( 'utf-8', errors = 'replace' )

    with open( file, 'wb' ) as fp:
      fp.write( src )

  return dump



import pytest
from copy import copy
from pprint import pprint

import itertools

from partis.schema import (
  required,
  optional,
  BoolPrim,
  IntPrim,
  FloatPrim,
  StrPrim,
  SeqPrim,
  MapPrim,
  UnionPrim,
  StructValued,
  PJCEvaluated,
  NotEvaluated,
  SchemaError,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  any_schema, )

from partis.schema.serialize.yaml import (
  loads,
  dumps )

from test_prim import (
  is_defined,
  get_val,
  should_pass,
  max_tests,
  step_tests )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# use this to only run sub-set of all possible tests
# max_tests = -1
# step_tests = 1

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def prim_union_cases( ):
  vals = [
    None,
    "$expr:py self",
    "$func:py return self",
    1,
    1.0,
    "abc",
    [ 1, 2, 3 ],
    { 'a': 'x', 'b': 'y', 'c': 'z' } ]

  class MySchema1( StructValued ):
    schema = dict(
      tag = 'my_schema1',
      struct = [
        ( 'a', StrPrim() ),
        ( 'b', StrPrim() ),
        ( 'c', StrPrim() ) ] )


  class MySchema2( StructValued ):
    schema = dict(
      tag = 'my_schema2',
      struct = [
        ( 'a', StrPrim() ),
        ( 'b', StrPrim() ),
        ( 'c', StrPrim() ) ] )


  valid_cases = [
    # valid combinations of cases
    ( True, [
      ( BoolPrim(), is_bool, None ),
      ( IntPrim(), is_numeric, None ),
      ( StrPrim(), is_string, None ),
      ( SeqPrim( item = IntPrim() ), is_sequence, is_numeric ),
      ( MapPrim( item = StrPrim() ), is_mapping, is_string ) ] ),
    ( True, [
      ( MySchema1, is_mapping, is_string ),
      ( MySchema2, is_mapping, is_string ) ] ),
    ( True, [
      ( BoolPrim(), is_bool, None ),
      ( FloatPrim(), is_numeric, None ),
      ( SeqPrim( item = IntPrim() ), is_sequence, is_numeric ),
      ( MySchema1, is_mapping, is_string ) ] ),
    # invalid combinations of cases
    ( False, [
      ( BoolPrim(), is_bool, None ),
      ( BoolPrim(), is_bool, None ) ] ),
    ( False, [
      ( IntPrim(), is_numeric, None ),
      ( FloatPrim(), is_numeric, None ) ] ),
    ( False, [
      ( IntPrim(), is_numeric, None ),
      ( IntPrim(), is_numeric, None ) ] ),
    ( False, [
      ( FloatPrim(), is_numeric, None ),
      ( FloatPrim(), is_numeric, None ) ] ),
    ( False, [
      ( StrPrim(), is_numeric, None ),
      ( StrPrim(), is_numeric, None ) ] ),
    ( False, [
      ( SeqPrim( item = IntPrim() ), is_sequence, is_numeric ),
      ( SeqPrim( item = StrPrim() ), is_sequence, is_string ) ] ),
    ( False, [
      ( MapPrim( item = IntPrim() ), is_mapping, is_numeric ),
      ( MapPrim( item = StrPrim() ), is_mapping, is_string )  ] ),
    ( False, [
      ( MapPrim( item = IntPrim() ), is_mapping, is_numeric ),
      ( MySchema1, is_mapping, is_string ) ] ) ]

  default_vals = [
    optional,
    required,
    True,
    1,
    "abc",
    [ 1, 2, 3 ],
    { 'a': 'x', 'b': 'y', 'c': 'z' },
    { 'type': 'my_schema1', 'a': 'x', 'b': 'y', 'c': 'z' },
    { 'type': 'my_schema2', 'a': 'x', 'b': 'y', 'c': 'z' } ]

  restricteds = [
    None ]

  evaluateds = [
    None,
    PJCEvaluated ]

  test_cases = itertools.product(
    vals,
    valid_cases,
    default_vals,
    restricteds,
    evaluateds )

  if max_tests >= 0:
    test_cases = itertools.islice( test_cases, 0, max_tests * step_tests, step_tests )

  return test_cases



#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
@pytest.fixture( params = prim_union_cases() )
def prim_union_case( request ):
  return request.param


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_prim_union( prim_union_case ):

  val, ( valid, cases ), default_val, restricted, evaluated = prim_union_case


  _should_pass_default_val = not is_defined( default_val )
  _should_pass_val = False
  num_schemas = sum( is_schema_struct(case) for case, is_type, is_item in cases )

  for case, is_type, is_item in cases:
    valid_val = True
    valid_default_val = True

    if is_schema_struct(case):
      # this checking the schema tag
      if is_mapping(val):
        if (
          num_schemas > 1
          and (
            case.tag_key not in val
            or val[case.tag_key] != case.tag ) ):

          valid_val = False

      if is_mapping(default_val):
        # print(type(case), case)
        # print(dir(case))

        if (
          num_schemas > 1
          and (
            case.tag_key not in default_val
            or default_val[case.tag_key] != case.tag ) ):

          valid_default_val = False

    _should_pass_val = (
      _should_pass_val
      or ( valid_val and should_pass(
        val = val,
        is_type = is_type,
        is_item = is_item,
        default_val = required if default_val is required else optional,
        restricted = restricted,
        evaluated = evaluated ) ) )

    _should_pass_default_val = (
      _should_pass_default_val
      or ( valid_default_val and should_pass(
        val = default_val,
        is_type = is_type,
        is_item = is_item,
        default_val = default_val,
        restricted = restricted,
        evaluated = evaluated ) ) )

    # print( _should_pass_val, _should_pass_default_val, case, is_type, is_item )


  _should_pass = _should_pass_val and _should_pass_default_val

  cases = [ c[0] for c in cases ]

  if valid and _should_pass:
    p = UnionPrim(
      cases = cases,
      default_val = default_val,
      restricted = restricted,
      evaluated = evaluated,
      doc = "My union" )

    schemas = [s for s in cases if is_schema_struct(s)]

    _default_val = copy(default_val)

    if is_mapping(default_val) and len(schemas) == 1 and schemas[0].tag_key in default_val:

      _default_val.pop(schemas[0].tag_key)

    assert p.default_val == _default_val
    assert p.restricted == restricted
    assert p.cases == cases
    assert p.doc.startswith("My union")
    assert p.evaluated == NotEvaluated if evaluated is None else evaluated

    compare = True
    _val = p.decode( val )
    p.encode( _val )

    _expected = get_val( val, default_val )

    if evaluated is not None and evaluated.check(val):
      if is_defined(default_val):
        # only check evaluated when default value is defined
        _val = _val._eval( context = { 'self' : default_val } )
        _expected = default_val
      else:
        compare = False

    if compare:
      if is_schema_struct_valued( _val ):
        for k, v in _expected.items():
          if k != _val._schema.tag_key:
            assert k in _val and v == _val[k]

      else:
        assert _expected == _val

  else:
    with pytest.raises(SchemaError):

      p = UnionPrim(
        cases = cases,
        default_val = default_val,
        restricted = restricted,
        evaluated = evaluated,
        doc = "My union" )

      _val = p.decode( val )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_prim_union_bias( ):

  a = BoolPrim(
    default_val = True )

  b = IntPrim(
    default_val = 55,
    min = 1,
    max = 100)

  d = StrPrim(
    default_val = 'abcd',
    pattern = r'.*',
    nonempty = True,
    max_lines = 3,
    max_cols = 100 )

  p = UnionPrim(
    cases = [a, b, d],
    default_case = 0 )

  v = p.decode()
  print(v._bias)
  assert float(v._bias) == 0

  v = p.decode(True)
  print(v._bias)
  assert float(v._bias) == 1

  v = p.decode(10)
  print(v._bias)
  assert float(v._bias) == 3

  v = p.decode('wxyz')
  print(v._bias)
  assert float(v._bias) == 8

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
if __name__ == '__main__':
  test_prim_union_bias()



import re
import inspect
from fnmatch import (
  translate )
import logging
log = logging.getLogger(__name__)

from math import log2
from collections.abc import (
  Mapping,
  Sequence )

from numbers import (
  Number,
  Complex,
  Real,
  Integral )

from rich.text import Text

from ruamel.yaml.comments import (
  CommentedBase,
  CommentedMap,
  CommentedSeq )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
from partis.utils.special import (
  SpecialType,
  RequiredType,
  OptionalType,
  DerivedType,
  NotSetType )

from partis.utils import (
  fmt_obj,
  hint_level_num,
  Loc,
  ModelHint,
  ModelError )

from partis.utils.inspect import (
  _filter_traceback )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import partis

LOG_TRACE = hint_level_num('trace')

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
name_re = r"^[a-zA-Z\_][a-zA-Z\_0-9]*$"
name_cre = re.compile( name_re )


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaError( ModelError ):
  """Base of all schema related errors
  """
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaNameError( SchemaError ):
  """Raised when a name error occurs
  """
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaParseError( SchemaError ):
  """Raised when a schema cannot be detected
  """
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaDetectionError( SchemaError ):
  """Raised when a schema cannot be detected
  """
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaDeclaredError( SchemaError ):
  """Raised when a schema's declared is used before the schema has been defined
  """
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaDefinitionError( SchemaError ):
  """Raised when a schema's definition is not correctly specified
  """
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaValidationError( SchemaError ):
  """Raised when data could not be validated (encoded/decoded) using schema definition
  """
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaEvaluationError( SchemaError ):
  """Raised when evaluated expression failed to result in a decoded value
  """
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaHint( ModelHint ):
  """Base of all schema hints
  """
  pass


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class Bias:
  """A metric of how biased (selective) a value is to a schema

  This may be interpreted as the number of needed independent binary decisions
  (questions, or conditional statements) to determine that the value satisfied
  the schema.

  Parameters
  ----------
  val : None | float
    The current/initial bias value
  const :  None | bool
    If ``True``, the bias value is considered to be a constant.

  """
  #-----------------------------------------------------------------------------
  def __init__(self, val = None, const = None ):
    if val is None:
      val = 0.0

    if isinstance(val, Bias):
      if const is None:
        const = val._const

      val = val._val

    self._const = const
    self._val = val

  #-----------------------------------------------------------------------------
  def __float__(self):
    return float(self._val)

  #-----------------------------------------------------------------------------
  def __add__(self, s):
    val = self._val

    if not self._const:
      val += float(s)

    return Bias(val, self._const)

  #-----------------------------------------------------------------------------
  def __iadd__(self, s):
    if not self._const:
      self._val += float(s)

    return self

  #-----------------------------------------------------------------------------
  def __str__(self):
    return str(float(self))

  #-----------------------------------------------------------------------------
  def __repr__(self):
    return f"{type(self).__name__}({float(self)})"

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class PresetValue:
  """A preset value associated with a schema

  Parameters
  ----------
  val : object
  label: None | str
  doc: None | str

  """
  #-----------------------------------------------------------------------------
  def __init__(self, val, label = None, doc = None ):
    if any(
      check( val )
      for check in [
        is_valued_type,
        is_evaluated ] ):

      val = val._encode

    if doc is None:
      doc = ""

    if label is None:
      label = fmt_obj(val, width = 10, height = 1)

    self._val = val
    self._label = label
    self._doc = doc

  #-----------------------------------------------------------------------------
  @property
  def val(self):
    return self._val

  #-----------------------------------------------------------------------------
  @property
  def label(self):
    return self._label

  #-----------------------------------------------------------------------------
  @property
  def doc(self):
    return self._doc

  #-----------------------------------------------------------------------------
  def __eq__(self, other):
    if isinstance(other, PresetValue):
      return self.val == other.val

    return self.val == other

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def assert_valid_name( name ):
  if not isinstance( name, str ):
    raise SchemaNameError(
      f"Name must be a string: {name}")

  match = name_cre.fullmatch(name)

  if match is None:
    raise SchemaNameError(
      f"Name must be only alpha-numeric or underscrores, and not start with a number: {name}")


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def assert_valid_path( path ):

  if isinstance( path, str ):
    if path == '':
      # empty path is valid
      return

    path = path.split('.')

  if not isinstance( path, list ) and all( isinstance(p, str) for p in path ):
    raise SchemaNameError(
      f"Name path must be a string or list of strings: {path}")

  for i, name in enumerate(path):
    try:

      assert_valid_name( name )

    except SchemaNameError as e:
      raise SchemaNameError(
        f"Path name segment {i} not valid: {path}",
        hints = SchemaHint.cast( e ) ) from e


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_special( val ):
  return isinstance( val, SpecialType )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_required( val ):
  """
  Returns
  -------
  bool
    True if the value corresponds to being required, likely resulting in
    an error if no value is supplied.
  """

  return isinstance( val, RequiredType )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_optional( val ):
  """
  Returns
  -------
  bool
    True if the value corresponds to being optional, resulting in a default of None
  """
  return val is None or isinstance( val, OptionalType )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_derived( val ):
  """
  Returns
  -------
  bool
    True if the value is to be derived from some other value(s)
  """
  return isinstance( val, DerivedType )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_notset( val ):
  """
  Returns
  -------
  bool
    True if the value corresponds to being not set, or otherwise undefined,
    including not even ``None``.
  """

  return isinstance( val, NotSetType )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_bool( obj ):
  """Is a boolean value
  """
  import partis.schema

  if isinstance( obj, type ):
    return issubclass( obj, (bool, partis.schema.BoolValued) )

  return isinstance( obj, (bool, partis.schema.BoolValued) )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_numeric( obj ):
  """Is a numeric value
  """

  if isinstance( obj, type ):
    return issubclass( obj, Number ) and not is_bool( obj )

  return isinstance( obj, Number ) and not is_bool( obj )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_string( obj ):
  """Is a string value
  """

  if isinstance( obj, type ):
    return issubclass( obj, (str, Text) )

  return isinstance( obj, (str, Text) )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_sequence( obj ):
  """Is a sequence value, but not a string
  """
  if isinstance( obj, type ):
    return issubclass( obj, Sequence ) and not is_string( obj )

  return isinstance( obj, Sequence ) and not is_string( obj )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_mapping( obj ):
  """Is a mapping value
  """
  if isinstance( obj, type ):
    return issubclass( obj, Mapping )

  return isinstance( obj, Mapping )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
type_tests = [
  is_required,
  is_derived,
  is_optional,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping ]

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_similar_value_type( a, b ):
  """
  Returns
  -------
  is_similar_value_type : bool
    True if all `type_tests` return the same for both values
  """
  return all( f(a) == f(b) for f in type_tests )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_schema( obj ):
  """Is a schema class or schema declared
  """
  return issubclass( type(obj), partis.schema_meta.schema.SchemaRef )


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_schema_prim( obj ):
  """Is a schema primitive type
  """
  return issubclass( type(type(obj)), partis.schema_meta.prim.SchemaPrimMeta )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_schema_declared( obj ):
  """Is a schema class or schema declared type
  """
  return issubclass( type(obj), partis.schema_meta.schema.SchemaDeclared )


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_schema_struct( obj ):
  """Is a schema class or schema declared type
  """
  return (
    issubclass( type(obj), partis.schema_meta.struct.SchemaStruct )
    or issubclass( type(obj), partis.schema_meta.struct.SchemaStructDeclared )
    or issubclass( type(obj), partis.schema_meta.struct.SchemaStructProxy ) )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_schema_struct_valued( obj ):
  """Is an instance value of a schema class type
  """
  return (
    is_valued_type( obj )
    and is_schema_struct(obj._schema) )


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_evaluated_class( val ):
  """Is an evaluated class type
  """
  return ( issubclass( type(val), partis.schema_meta.eval.EvaluatedMeta ) )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_provider( val ):
  """Is an evaluated type
  """
  return issubclass( type(type(val)), partis.schema_meta.eval.ProviderMeta )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_evaluated( val ):
  """Is an evaluated type
  """
  if is_valued_type(val):
    return is_evaluated(val._src)

  return issubclass( type(type(val)), partis.schema_meta.eval.EvaluatedMeta )


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_valued_type( val ):
  """
  Returns
  -------
  is_valued_type : bool
    True if is an instance value of any schema type
  """
  return issubclass( type(type(val)), partis.schema_meta.valued.ValuedMeta )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def any_schema( val, schemas ):
  """Is the schema any of the list schemas

  Parameters
  ----------
  schema : object
    Schema to test
  schema : list[object]
    Schemas to compare to
  """
  if is_valued_type( val ):
    return any( val._schema is s.schema for s in schemas )

  elif is_schema( val ):
    return any( val.schema is s.schema for s in schemas )

  return False


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def is_valued( val ):
  """
  Returns
  -------
  is_valued : bool
    True if value is completely evaluated
  """

  if is_evaluated(val):
    return False

  if is_valued_type( val ):
    return val._valued

  if is_sequence( val ):
    return all( is_valued(v) for v in val )

  if is_mapping( val ):
    return all( is_valued(v) for v in val.values() )

  return True

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def get_line_col(obj, key = None):

  if isinstance( obj, CommentedBase ):
    # NOTE: ruamel appears to store line/col in zero-based indexing
    if (
      key is None
      or not ( isinstance(obj, CommentedMap) or isinstance(obj, CommentedSeq) )
      or obj.lc.data is None
      or (isinstance(obj, CommentedMap) and key not in obj)
      or (isinstance(obj, CommentedSeq) and ( key < 0 or key >= len(obj) ) ) ):

      return obj.lc.line + 1, obj.lc.col + 1

    else:
      return obj.lc.data[key][0] + 1, obj.lc.data[key][1] + 1

  if obj is not None:
    try:
      return inspect.getsourcelines(obj)[1], None
    except:
      pass

  return None, None

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class schema_errors:

  #-----------------------------------------------------------------------------
  def __init__(self,
    msg = None,
    data = None,
    loc = None,
    hints = None,
    schema = None,
    cls = None,
    filter = None,
    suppress = False ):

    if cls is None:
      cls = SchemaError

    if msg is None:
      msg = f"Error while validating"

    if hints is None:
      hints = list()

    if not isinstance(hints, list):
      hints = [hints]

    if schema is not None:
      if loc is None:
        loc = Loc( owner = schema )
      else:
        loc.owner = schema

    assert issubclass(cls, SchemaError)

    patterns = [r'*/site-packages/partis/*']
    patterns = [
      re.compile( translate(p) )
      for p in patterns ]

    def _ignore(fname):
      return any(p.match(fname) is not None for p in patterns)

    if filter is None:
      filter = (Exception,)
    elif not isinstance(filter, (list, tuple)):
      filter = (filter,)
    else:
      filter = tuple(filter)

    self._cls = cls
    self._ignore = _ignore
    self._filter = filter
    self._suppress = bool(suppress)

    self.msg = msg
    self.data = data
    self.loc = loc
    self.hints = hints

  #-----------------------------------------------------------------------------
  def __enter__(self):
    return self

  #-----------------------------------------------------------------------------
  def __exit__(self, cls, value, traceback):

    if cls is not None and issubclass(cls, self._filter):
      if not log.isEnabledFor(LOG_TRACE):
        _filter_traceback(
          traceback,
          self._ignore,
          keep_last = not (hasattr(value, 'ignore_frame') and value.ignore_frame))

      msg = self.msg

      raise self._cls(
        msg,
        data = self.data,
        loc = self.loc,
        hints = self.hints + [value],
        ignore_frame = True ) from None

    # do not handle any exceptions here
    return self._suppress



# -*- coding: UTF-8 -*-

import logging
log = logging.getLogger(__name__)

from copy import copy
import inspect
import re

from partis.utils import (
  odict,
  fmt_base_or_type,
  fmt_iterable_or,
  fmt_attr_doc )

from partis.pyproj import (
  hash_sha256 )

from partis.schema_meta.base import (
  SchemaError,
  SchemaNameError,
  SchemaDeclaredError,
  SchemaDefinitionError,
  SchemaValidationError,
  SchemaHint,
  Bias,
  Loc,
  assert_valid_name,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  is_evaluated,
  is_valued,
  is_optional,
  is_required )

from partis.schema.valued import (
  StrValued )

from . import (
  SchemaPrimDeclared,
  SchemaPrim )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class StrPrimDeclared( SchemaPrimDeclared ):
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class StrPrim( SchemaPrim ):
  """Primitive for string values

  Parameters
  ----------
  char_case : NoneType | str
    Cased characters are converted to given case

    - None (default) : Cased characters are not altered

    - 'lower' : Cased characters are converted to their lower-case equivalent

    - 'upper' : Cased characters are converted to their upper-case equivalent
  strip : NoneType | bool
    Strips leading and trailing white-space if ``True``.
  pattern : NoneType | str
    Validates a string using a regular expression (Python syntax)

    .. note::

      The pattern is only matched to non-empty strings.
      See the `nonempty` argument to only allow string that have at least one character.
  nonempty : NoneType | bool
    Only non-empty strings are valid if ``True``.
    If the value is optional, the no-value or a value of ``None`` is still considered valid.
  max_lines : NoneType | int
    Maximum number of lines
  max_cols : NoneType | int
    Maximum number of columns

  **kwargs : arguments passed to :class:`SchemaPrim <partis.schema.prim.base.SchemaPrim>`
  """

  #-----------------------------------------------------------------------------
  def __new__( mcls,
    name,
    bases,
    namespace,
    char_case = None,
    strip = None,
    pattern = None,
    nonempty = None,
    max_lines = None,
    max_cols = None,
    valued_type = None,
    **kwargs ):

    loc = kwargs.get('loc', None)

    lines = list()

    if char_case not in [ None, 'lower', 'upper' ]:
      raise SchemaDefinitionError(
        f"`char_case` must be None, lower, or upper: {char_case}",
        loc = loc )

    strip = bool(strip)

    if pattern is not None:
      try:
        pattern = re.compile( pattern )

      except BaseException as e:
        raise SchemaDefinitionError(
          f"`pattern` must be a valid Python regular expression: {pattern}",
          loc = loc ) from e

    nonempty = bool(nonempty)

    if max_lines is not None:
      max_lines = int(max_lines)

      if max_lines < 1:
        raise SchemaDefinitionError(
          f"`max_lines` must be >= 1: {max_lines}",
          loc = loc )

    if max_cols is not None:
      max_cols = int(max_cols)

      if max_cols < 1:
        raise SchemaDefinitionError(
          f"`max_cols` must be >= 1: {max_cols}",
          loc = loc )

    if valued_type is None:
      valued_type = StrValued

    namespace = { **namespace, **dict(
      _p_char_case = char_case,
      _p_strip = strip,
      _p_pattern = pattern,
      _p_nonempty = nonempty,
      _p_max_lines = max_lines,
      _p_max_cols = max_cols )}

    return super().__new__( mcls,
      name = name,
      bases = bases,
      namespace = namespace,
      valued_type = valued_type,
      **kwargs )

  #-----------------------------------------------------------------------------
  def __init__( cls,
    name,
    bases,
    namespace,
    char_case = None,
    strip = None,
    pattern = None,
    nonempty = None,
    max_lines = None,
    max_cols = None,
    **kwargs ):


    super().__init__(
      name = name,
      bases = bases,
      namespace = namespace,
      **kwargs )

  #-----------------------------------------------------------------------------
  def _get_attr_doc_lines( cls,
    noindex = False,
    depth = 0,
    max_depth = 2 ):

    lines = super()._get_attr_doc_lines(
      noindex = noindex,
      depth = depth,
      max_depth = max_depth )

    lines.extend([
      fmt_attr_doc(
        name = attr,
        typename = type( getattr( cls, attr ) ),
        val = getattr( cls, attr ),
        prefix = 'schema',
        noindex = noindex )
      for attr in [
        'char_case',
        'strip',
        'pattern',
        'nonempty',
        'max_lines',
        'max_cols' ] ])

    return lines

  #-----------------------------------------------------------------------------
  def _schema_hash_node( cls ):
    lines = super()._schema_hash_node()

    for attr in [
      'char_case',
      'strip',
      'pattern',
      'nonempty',
      'max_lines',
      'max_cols' ]:

      val = getattr( cls, attr )
      hash = None

      if hasattr( val, 'schema_hash' ):
        hash = val.schema_hash
      else:
        hash = hash_sha256( str(val).encode('utf-8') )[0]

      lines.append( f'{attr}, {hash}' )

    return lines

  #-----------------------------------------------------------------------------
  @property
  def char_case( cls ):
    return cls._p_char_case

  #-----------------------------------------------------------------------------
  @property
  def pattern( cls ):
    return cls._p_pattern

  #-----------------------------------------------------------------------------
  @property
  def max_lines( cls ):
    return cls._p_max_lines

  #-----------------------------------------------------------------------------
  @property
  def max_cols( cls ):
    return cls._p_max_cols

  #-----------------------------------------------------------------------------
  @property
  def nonempty( cls ):
    return cls._p_nonempty

  #-----------------------------------------------------------------------------
  @property
  def strip( cls ):
    return cls._p_strip

  #-----------------------------------------------------------------------------
  def validate( cls,
    val,
    loc = None,
    bias = None ):

    bias = Bias(bias)

    #...........................................................................
    if not is_string( val ):
      raise SchemaValidationError(
        f"Must be a `str`: {type(val).__name__}",
        loc = loc )

    bias += 1
    val = str(val)

    #...........................................................................
    if cls.char_case is not None:
      if cls.char_case == 'lower':
        val = str.lower( val )

      elif cls.char_case == 'upper':
        val = str.upper( val )

      else:
        assert False

    #...........................................................................
    if cls.strip:
      val = val.strip()

    #...........................................................................
    if cls.pattern is not None:
      bias += len(val)

      if len(val) > 0 and cls.pattern.fullmatch( val ) is None:
        raise SchemaValidationError(
          f"Does not match pattern `{cls.pattern.pattern}`:",
          data = val,
          loc = loc )

    if cls.nonempty:
      bias += 1

      if len(val) == 0:
        raise SchemaValidationError(
          f"Must be non-empty string",
          loc = loc )

    #...........................................................................
    if cls.max_lines is not None or cls.max_cols is not None:
      lines = val.splitlines()

    #...........................................................................
    if cls.max_lines is not None:
      bias += 1

      if len(lines) > cls.max_lines:
        raise SchemaValidationError(
          f"Maximum lines exceeded {cls.max_lines}: {len(lines)}",
          loc = loc )

    #...........................................................................
    if cls.max_cols is not None:
      cols = list( map(len, lines) )
      bias += len(cols)

      if len(cols) > 0:
        max_cols = max(cols)

        if max_cols > cls.max_cols:
          idx = cols.index( max_cols )

          raise SchemaValidationError(
            f"Maximum columns exceeded {cls.max_cols}: {max_cols} (line {idx})",
            loc = loc )

    return super().validate(
      val = val,
      loc = loc,
      bias = bias )



# -*- coding: UTF-8 -*-

import os

try:
  from importlib.metadata import distributions

except ImportError:
  from importlib_metadata import distributions

from partis.utils import getLogger
log = getLogger(__name__)

from partis.utils.plugin import (
  Plugin,
  PluginError,
  plugin_manager )

from partis.schema import (
  is_string,
  is_sequence,
  is_schema )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def get_base_schemas():
  from .hint import (
    HintList,
    HintLoc,
    Hint )

  return [ SchemaPluginGroup(
    label = "",
    schemas = {
      'Hints' : HintList }) ]

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaPluginError( PluginError ):
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaPluginGroup( Plugin ):
  """Group of schemas to load as a plugin

  Parameters
  ----------
  schemas : dict[str, Schema]
    Dictionary of loaded schemas. Keys are the labels for each schema in the
    group.
  label : str
    User-friendly label for the group of schemas loaded by this plugin
  """

  #-----------------------------------------------------------------------------
  def __init__( self,
    schemas,
    label = None ):

    if label is None:
      label = ''

    for k, v in schemas.items():
      if not is_string(k):
        raise SchemaPluginError(
          f'Key must be a string: {type(k)}' )

      if not is_schema(v):
        raise SchemaPluginError(
          f'Value must be a schema: {type(v)}' )

      # ensure schemas are not a proxy
      schemas[k] = v.schema


    self._p_label = str(label)
    self._p_schemas = dict(schemas)

  #-----------------------------------------------------------------------------
  @property
  def label(self):
    return self._p_label

  #-----------------------------------------------------------------------------
  @property
  def schemas(self):
    return self._p_schemas

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaPluginManager:
  """Manages schema plugins

  Distributions providing schema plugins should define an 'entry_point' with
  a group name of 'partis.schema' that is a callable returning a list of
  instances of :class:`SchemaPluginGroup`.
  The loaded schemas can be used for automatic discovery of schemas to use for
  decoding contents using the .
  """
  #-----------------------------------------------------------------------------
  def __init__( self ):

    self._hash_lookup = dict()

  #-----------------------------------------------------------------------------
  def get_by_hash( self,
    hash ):
    """Find schema by the schema_hash

    Parameters
    ----------
    hash : str
      Value corresponding to that returned by a schema's 'schema_hash' property.

    Returns
    -------
    list[Schema]
      List of schemas matching the hash, or an empty list if none are found.
    """

    self.ensure_loaded()

    if hash not in self._hash_lookup:
      return list()

    return self._hash_lookup[ hash ]

  #-----------------------------------------------------------------------------
  def register_plugin( self,
    plugin ):

    if not isinstance( plugin, SchemaPluginGroup ):
      raise SchemaPluginError(
        f'plugin must be SchemaPluginGroup: {type(plugin)}')

    for schema in plugin.schemas.values():
      hash = schema.schema_hash

      if hash not in self._hash_lookup:
        self._hash_lookup[ hash ] = list()

      self._hash_lookup[ hash ].append( schema )

  #-----------------------------------------------------------------------------
  def ensure_loaded( self ):
    plugin_manager.ensure_loaded()

  #-----------------------------------------------------------------------------
  def load_plugins( self ):

    plugin_manager.load_plugins()

    for plugin in plugin_manager.plugins( SchemaPluginGroup ):
      self.register_plugin( plugin )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
schema_plugins = SchemaPluginManager()



# -*- coding: UTF-8 -*-

import logging
log = logging.getLogger(__name__)

from copy import copy
import inspect
import re

from partis.utils import (
  odict,
  fmt_base_or_type,
  fmt_iterable_or )

from partis.schema_meta.base import (
  SchemaError,
  SchemaNameError,
  SchemaDeclaredError,
  SchemaDefinitionError,
  SchemaValidationError,
  SchemaHint,
  Loc,
  assert_valid_name,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  is_evaluated,
  is_valued,
  is_optional,
  is_required )

from partis.schema_meta.schema import (
  PassValued )

from . import (
  SchemaPrim )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class PassPrim( SchemaPrim ):
  """Primitive for an un-checked (pass-through) value

  Parameters
  ----------
  **kwargs : arguments passed to :class:`SchemaPrim <partis.schema.prim.SchemaPrim>`
  """




from copy import copy
import io
import re
import uuid
import inspect
import sys
import importlib

from abc import ABCMeta
import warnings

import logging
log = logging.getLogger(__name__)

from partis.utils import (
  cached_property,
  adict,
  rdict,
  rlist,
  StringFunction,
  mapping_attrs,
  protected_attr,
  indent_lines,
  split_lines,
  fmt_obj,
  fmt_base_or_type,
  _fmt_class_name,
  fmt_class_name,
  fmt_attr_doc,
  make_dynamic_class_name )

from partis.pyproj import (
  hash_sha256 )

from partis.utils.special import (
  required,
  optional,
  derived,
  notset )

from partis.schema_meta.base import (
  schema_errors,
  SchemaError,
  SchemaNameError,
  SchemaDeclaredError,
  SchemaDefinitionError,
  SchemaValidationError,
  SchemaHint,
  Loc,
  assert_valid_name,
  assert_valid_path,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_special,
  is_optional,
  is_required,
  is_derived,
  is_notset,
  is_similar_value_type,
  is_schema_prim,
  is_schema_declared,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  is_evaluated_class,
  is_evaluated,
  is_valued,
  is_valued_type,
  any_schema )

from partis.schema_meta.property import (
  SchemaProperty,
  ConstProperty )

from .schema import (
  fmt_schema_typename,
  SchemaProxy,
  SchemaDeclared,
  Schema )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaStructDeclared( SchemaDeclared ):
  """Meta-class for all schema declareds

  A schema declared acts as a forward declaration of a schema class if a reference
  to the schema is needed before the schema class has actually been defined.
  The eventual schema definition must have the exact same `tag_key` and `tag`
  as the declared that it defines.
  Once it has been defined, the actual schema definition may be retrieved from the
  `schema` property.
  A declared may only be defined by a *single* eventual schema class.

  Parameters
  ----------
  name : str
    Class name of type
  bases : list[type]
    Base classes of type
  namespace : dict[str, object]
  tag_key : NoneType | str
    The key for the tag defining this schema.
    (default: 'type')
  tag : str
    The tag defining this schema
  See Also
  --------
  :class:`SchemaDeclared <partis.schema.declared.SchemaDeclared>`
  :class:`SchemaStruct <partis.schema.struct.SchemaStruct>`

  """

  #-----------------------------------------------------------------------------
  def __new__( mcls,
    name,
    bases,
    namespace,
    tag,
    tag_key = None ):


    if tag_key is None:
      tag_key = 'type'

    try:
      assert_valid_name( tag_key )
      assert_valid_name( tag )

    except SchemaNameError as e:

      raise SchemaDeclaredError(
        f"Schema declared name not valid",
        hints = e ) from e

    namespace = {**namespace, **dict(
      _p_tag_key = tag_key,
      _p_tag = tag )}

    cls = super().__new__( mcls, name, bases, namespace )

    return cls

  #-----------------------------------------------------------------------------
  def __init__( cls,
    name,
    bases,
    namespace,
    tag,
    tag_key = None ):

    super().__init__( name, bases, namespace )


  #-----------------------------------------------------------------------------
  @property
  def tag_key( cls ):
    """str : The key for the tag defining this schema
    """
    return cls._p_tag_key


  #-----------------------------------------------------------------------------
  @property
  def tag( cls ):
    """str : The tag defining this schema
    """
    return cls._p_tag


  #-----------------------------------------------------------------------------
  def _schema_declared( cls, schema ):
    schema = super()._schema_declared( schema )

    if not is_schema_struct(schema):
      raise SchemaDeclaredError(
        f"Must be declared for a SchemaStruct" )

    if (
      schema.tag != cls.tag or
      schema.tag_key != cls.tag_key ):

      raise SchemaDeclaredError(
        f"Defining schema declared must have the same `tag_key` '{cls.tag_key}' and `tag` '{cls.tag}'",
        hints = schema.hints )

    return schema

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaStruct( Schema ):
  """Meta-class for all defined schemas

  Parameters
  ----------
  name : str
    Class name of type
  bases : list[type]
    Base classes of type
  namespace : dict[str, object]
  tag_key : NoneType | str
    The key for the tag defining this schema
  tag : NoneType | str
    The tag defining this schema
  declared : NoneType | :class:`SchemaDeclared <partis.schema.declared.SchemaDeclared>`
    The schema declared for which this is the definition.
    If given, the `tag_key`, `tag` are not required.
  struct : NoneType | dict[ str, :class:`SchemaStruct <partis.schema.struct.SchemaStruct>` | :class:`SchemaPrim <partis.schema.prim.base.SchemaPrim>` ]
    Structure of the defined schema Mapping.
    Keys must be valid attribute names, not be "private" names starting with
    '__' (double-underscore) or '_p', or names of any Mapping attributes 'keys',
    'values', 'items', or 'get'.
  struct_proxy : NoneType | :class:`RequiredType <partis.utils.special.RequiredType>` | str
    Key of primary structure value that may be set, using defaults for all other
    values.
    If any other values are required, specifying this will raise an error.

  Raises
  ------
  SchemaDefinitionError
    If the schema definition is not valid

  See Also
  --------
  :class:`SchemaDeclared <partis.schema.declared.SchemaDeclared>`
  :class:`SchemaStruct <partis.schema.struct.SchemaStruct>`
  """

  #-----------------------------------------------------------------------------
  def __new__(mcls,
    name,
    bases,
    namespace,
    tag_key = None,
    tag = None,
    declared = None,
    struct = None,
    struct_proxy = None,
    evaluated = None,
    default_val = None,
    default_eval = None,
    init_val = None,
    preset_vals = None,
    valued_type = None,
    doc = None,
    loc = None ):

    if default_val is None:
      default_val = required

    if default_eval is None:
      default_eval = required

    if declared is not None:
      # defines the given schema declared

      if not isinstance( declared, SchemaStructDeclared ):
        raise SchemaDefinitionError(
          f"`declared` must be subclass of SchemaStructDeclared: {fmt_base_or_type(declared)}",
          loc = loc )

      if tag_key is not None and tag_key != declared.tag_key:
        raise SchemaDefinitionError(
          f"declared `tag_key` {declared.tag_key} redefined: {tag_key}",
          loc = loc )

      if tag is not None and tag != declared.tag:
        raise SchemaDefinitionError(
          f"declared `tag` {declared.tag} redefined: {tag}",
          loc = loc )

      tag_key = declared.tag_key
      tag = declared.tag

    if tag_key is None:
      tag_key = 'type'


    if tag is None:
      tag = 'base'

    try:
      assert_valid_name( tag_key )
      assert_valid_name( tag )

    except SchemaNameError as e:
      raise SchemaDefinitionError(
        f"Schema name not valid",
        loc = loc,
        hints = SchemaHint.cast( e ) ) from e


    if struct is None:
      struct = dict()

    if not isinstance( struct, dict ):
      if is_sequence( struct ) or is_mapping( struct ):
        try:
          struct = dict( struct )

        except BaseException as e:
          raise SchemaDefinitionError(
            f"Schema struct must be an ordered dictionary, or items list: {struct}",
            loc = loc,
            hints = e )

      else:
        raise SchemaDefinitionError(
          f"Schema struct must be a mapping or sequence of key-value pairs: {struct}",
          loc = loc )

    else:
      struct = copy(struct)


    schema_deps = list()

    # validate struct items
    if tag_key in struct:
      raise SchemaDefinitionError(
        f"Schema `struct` may not define the key used as the schema's `tag_key`: {tag_key}",
        loc = loc )

    for k, v in struct.items():
      if not is_string( k ):
        raise SchemaDefinitionError(
          f"Schema struct key must be string: {k}",
          loc = loc )

      if any( k.startswith(p) for p in protected_attr ):
        raise SchemaDefinitionError(
          f"Key may not start with {protected_attr}: {k}",
          loc = loc )

      if k in mapping_attrs:
        raise SchemaDefinitionError(
          f"Key may not be of any 'Mapping' type attributes {mapping_attrs}: {k}",
          loc = loc )

      try:
        assert_valid_name( k )

      except SchemaNameError as e:
        raise SchemaDefinitionError(
          f"Schema struct key not valid",
          loc = loc,
          hints = e ) from e

      if not is_schema(v):
        raise SchemaDefinitionError(
          f"Schema struct value '{k}' must be subclass of `SchemaStruct` or `SchemaPrim`: {type(v)}",
          loc = loc )

      if not v.schema_defined:
        schema_deps.append(v)

    if evaluated is None:
      from partis.schema.eval import NotEvaluated
      evaluated = NotEvaluated

    if not is_evaluated_class(evaluated):
      raise SchemaDefinitionError(
        f"`evaluated` must be subclass of `Evaluated`: {fmt_base_or_type(evaluated)}",
        loc = loc )

    if is_required(struct_proxy):
      struct_proxy = None

    # evaluated_check = evaluated.check if evaluated else lambda x: False
    if struct_proxy:

      if not is_string(struct_proxy):
        raise SchemaDefinitionError(
          f"Schema `struct_proxy` key must be a string value: {struct_proxy}",
          loc = loc )

      if struct_proxy not in struct:
        raise SchemaDefinitionError(
          f"Schema `struct_proxy` key must be defined in schema: {struct_proxy}",
          loc = loc )

    #...........................................................................
    # for k, v in struct.items():
    #   namespace[k] = SchemaProperty(
    #     schema = v,
    #     name = k )
    #
    # namespace[ tag_key ] = ConstProperty(
    #   name = tag_key )

    namespace = { **namespace, **dict(
      _p_tag_key = tag_key,
      _p_tag = tag,
      _p_struct = struct,
      _p_struct_proxy = struct_proxy,
      _p_declared = declared )}

    cls = super().__new__(
      mcls,
      name,
      bases,
      namespace,
      evaluated = evaluated,
      default_val = default_val,
      default_eval = default_eval,
      init_val = init_val,
      preset_vals = preset_vals,
      schema_deps = schema_deps,
      valued_type = valued_type,
      doc = doc,
      loc = loc )

    return cls

  #-----------------------------------------------------------------------------
  def __init__( cls,
    name,
    bases,
    namespace,
    tag_key = None,
    tag = None,
    declared = None,
    struct = None,
    struct_proxy = None,
    evaluated = None,
    default_val = None,
    default_eval = None,
    init_val = None,
    preset_vals = None,
    valued_type = None,
    doc = None,
    loc = None ):

    # super().__init__( name, bases, namespace )

    if cls._p_declared is not None:
      cls._p_declared.schema_declared( cls )

    super().__init__( name, bases, namespace )

  #-----------------------------------------------------------------------------
  def _get_attr_doc_lines( cls,
    noindex = False,
    depth = 0,
    max_depth = 2 ):

    lines = super()._get_attr_doc_lines(
      noindex = noindex,
      depth = depth,
      max_depth = max_depth )

    lines.extend([
      fmt_attr_doc(
        name = attr,
        typename = type( getattr( cls, attr ) ),
        obj = cls,
        val = getattr( cls, attr ),
        prefix = 'schema',
        # don't index these attributes on the instance of the meta-class, since
        # these are actually metaclass attributes
        noindex = True )
      for attr in [
        'tag_key',
        'tag',
        'struct_proxy',
        'default_val',
        'default_eval',
        'init_val' ] ])


    lines.append( fmt_attr_doc(
      name = 'evaluated',
      typename = cls.evaluated,
      obj = cls,
      prefix = 'schema',
      noindex = noindex ) )

    lines.extend([
      fmt_attr_doc(
        name = k,
        typename = fmt_schema_typename(v),
        obj = cls,
        prefix = 'schema.struct',
        noindex = noindex,
        doc = v._get_doc(
          # always specify noindex since the item schema already documented
          noindex = True,
          depth = depth + 1,
          max_depth = max_depth ) )
      for k, v in cls.struct.items() ])

    return lines

  #-----------------------------------------------------------------------------
  def _schema_hash_node( cls ):

    lines = super()._schema_hash_node()

    for attr in [
      'tag_key',
      'tag',
      # TODO: should these be included in the hash? Default values used while
      # decoding seem like non-essential information for validating what was
      # originally encoded.
      # 'default_val',
      # 'default_eval',
      # 'init_val',
      'evaluated',
      'struct_proxy' ]:

      val = getattr( cls, attr )
      hash = None

      if hasattr( val, 'schema_hash_node' ):
        hash = val.schema_hash_node
      else:
        hash = hash_sha256( str(val).encode('utf-8') )[0]

      lines.append( f'{attr}, {hash}' )

    lines.append( 'keys, ' + str(list(cls.struct.keys())) )

    return lines

  #-----------------------------------------------------------------------------
  def _schema_hash_children( cls ):
    return super()._schema_hash_children() + list(cls.struct.values())

  #-----------------------------------------------------------------------------
  def _schema_resolved( cls ):
    super()._schema_resolved()

    if cls.struct_proxy:
      for k, v in cls.struct.items():
        v = v.schema

        if k != cls.struct_proxy and is_required( v.default_val ):
          # NOTE: this only validates that no other items in the struct are required

          raise SchemaDefinitionError(
            f"Schema with a `struct_proxy` may not contain other required values: {k}",
            loc = cls.loc,
            hints = v.hints )

    return True

  #-----------------------------------------------------------------------------
  @property
  def tag_key( cls ):
    """str : Key used to obtain the tag for this struct type
    """
    return cls._p_tag_key

  #-----------------------------------------------------------------------------
  @property
  def tag( cls ):
    """str : Tag for this struct type
    """
    return cls._p_tag

  #-----------------------------------------------------------------------------
  @property
  def struct( cls ):
    """dict[ str, :class:`SchemaStruct <partis.schema.struct.SchemaStruct>` | :class:`SchemaPrim <partis.schema.prim.SchemaPrim>` ] : Structure
    of the defined schema
    """
    return cls._p_struct

  #-----------------------------------------------------------------------------
  @property
  def struct_proxy( cls ):
    return cls._p_struct_proxy

  #-----------------------------------------------------------------------------
  @cached_property
  def default_val( cls ):

    val = cls._p_default_val

    if not (
      is_optional( val )
      or is_required( val )
      or is_derived( val ) ):

      val = copy(val)

    if cls.struct_proxy:

      # standardize values given the struct_proxy
      if not any(
        check( val )
        for check in [
          is_required,
          is_optional,
          is_derived,
          cls.evaluated.check,
          is_mapping ] ):

        val = { cls.struct_proxy : val }

    if is_derived( val ):
      # attempt to derive a default value

      val = { cls.tag_key : cls.tag }

      for k, _schema in cls.struct.items():
        _schema = _schema.schema

        _val = _schema.default_val

        if is_required( _val ):
          raise SchemaDefinitionError(
            f"Cannot derive default from required values: {k}",
            loc = cls.loc)

        if is_valued_type( _val ):
          #TODO is this needed? all default_val should return basic types
          _val = _val._encode

        if not is_optional( _val ):
          if is_mapping( _val ) and is_schema_struct( _schema ):
            # don't need to store tag key since schema is not ambiguous
            if _schema.tag_key in _val:
              _val.pop( _schema.tag_key )

          val[k] = _val

    # ensure default value includes the tag_key: tag
    if is_mapping(val) and cls.tag_key not in val:
      val[ cls.tag_key ] = cls.tag

    return val

    # TODO: why was this here?
    # if is_optional(val) or is_required(val):
    #   return val
    # else:
    #   return cls.encode( val, cls.loc )

  #-----------------------------------------------------------------------------
  @cached_property
  def default_eval( cls ):

    val = cls._p_default_eval

    if cls.struct_proxy:

      # standardize values given the struct_proxy
      if not any(
        check( val )
        for check in [
          is_required,
          is_mapping ] ):

        val = { cls.struct_proxy : val }

    if is_required(val):
      return val
    else:
      return cls.encode( val, cls.loc )

  #-----------------------------------------------------------------------------
  @cached_property
  def init_val( cls ):
    val = cls._p_init_val

    if val is not None:
      if cls.struct_proxy:
        if not any(
          check( val )
          for check in [
            cls.evaluated.check,
            is_mapping ] ):

          val = { cls.struct_proxy : val }


      return cls.encode( val, cls.loc )

    # default_eval checked first, only not-required if `default_val` is an expression

    val = cls.default_eval

    if not is_required( val ):
      return val

    val = cls.default_val

    if not any(
      check( val )
      for check in [
        is_required,
        is_optional,
        is_derived,
        cls.evaluated.check ] ):

      # only use `default_val` if an actual value, and not an expression
      return val


    # attempt to derive an initial value
    val = { cls.tag_key : cls.tag }

    for k, _schema in cls.struct.items():
      _schema = _schema.schema

      if not is_optional( _schema.default_val ):
        _val = _schema.init_val

        val[k] = _val

    return cls.encode( val, cls.loc )

  #-----------------------------------------------------------------------------
  @property
  def evaluated( cls ):
    return cls._p_evaluated

  #-----------------------------------------------------------------------------
  @cached_property
  def hints( cls ):

    # build hints
    # NOTE: hints are constructed dynamically because initially some schemas
    # may not be fully defined with 'declared' placeholders.

    hints = super().hints

    hints.append( SchemaHint(
      f"{cls.tag_key}: {cls.tag}" ) )

    hints.append( SchemaHint(
      f"struct_proxy: {cls.struct_proxy}" ) )

    # NOTE: must check default value here in case it was a derived value
    default_val = cls.default_val

    if is_required( default_val ):
      hints.append( SchemaHint(
        f"default_val: required" ) )

    elif is_optional( default_val ):
      hints.append( SchemaHint(
        f"default_val: optional" ) )

    else:

      hints.append( SchemaHint(
        "default_val: dict",
        hints = [ default_val ] ) )

    if len(cls.struct) == 0:
      struct_hints = "Empty schema structure"

    else:
      struct_hints = list()

      for k, v in cls.struct.items():
        if is_schema_declared(v):
          struct_hints.append( SchemaHint( f"{k}" ) )

        else:
          struct_hints.append( SchemaHint(
            f"{k}:",
            hints = v.hints ) )


    hints.append( SchemaHint(
      "struct: dict[str, schema]",
      hints = struct_hints ) )


    return hints

  #-----------------------------------------------------------------------------
  def encode( cls,
    val = None,
    loc = None,
    no_defaults = None ):

    if is_valued_type( val ):

      if loc is None:
        loc = val._loc

      if val._schema.schema is not cls or is_evaluated( val._src ):
        val = val._encode

    if loc is None:
      loc = Loc(path = [cls.__name__])

    if is_optional(val):
      val = cls.default_val

      if is_required( val ):
        raise SchemaValidationError(
          f"Value is required",
          loc = loc,
          hints = cls.doc )

      elif is_optional( val ):
        return None

    if cls.evaluated.check( val ):
      return val

    if not is_mapping( val ):
      if cls.struct_proxy:
        # uses source data for primary struct key value with remaining defaults
        val = { cls.struct_proxy : val }

      else:
        raise SchemaValidationError(
          f"Value must be a mapping: {type(val).__name__}",
          loc = loc )

    _val = {
      cls.tag_key : cls.tag }

    for k, _schema in cls.struct.items():
      _schema = _schema.schema

      v = val.get(k)
      _loc = loc(key = k)

      with schema_errors(
        msg = f"Encode failed",
        loc = _loc,
        schema = cls,
        cls = SchemaValidationError ):

        v = _schema.encode(
          val = v,
          loc = _loc,
          no_defaults = no_defaults )

      if v is not None:
        if no_defaults and (
          (
            v == _schema.default_val
            and is_similar_value_type(v, _schema.default_val) )
          or (
            is_mapping( v )
            and is_schema_struct( _schema )
            and len(v) == 1
            and _schema.tag_key in v ) ):

          pass

        else:
          if is_mapping( v ) and is_schema_struct( _schema ):
            # don't need to store tag key since schema is not ambiguous
            v.pop( _schema.tag_key )

          _val[k] = v

    val = _val

    if no_defaults and (
      len(val) == 2
      and cls.struct_proxy
      and cls.struct_proxy in val ):

      val = val[cls.struct_proxy]

    # if no_defaults and (
    #   # one item when only the tag_key: tag exist
    #   len(val) == 1
    #   # check that equivalent to default value
    #   or (
    #     val == cls.default_val
    #     and all(
    #       is_similar_value_type(a,b)
    #       for a,b in zip( val.values(), cls.default_val.values() )) )):
    #
    #   val = None

    return val

  #-----------------------------------------------------------------------------
  def decode( cls,
    val = None,
    loc = None,
    bias = None ):

    valued_type = cls.valued_type

    return valued_type( val, loc, bias )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class SchemaStructProxy( SchemaProxy ):

  #-----------------------------------------------------------------------------
  def __new__(mcls,
    name,
    bases,
    namespace,
    tag_key = None,
    tag = None,
    declared = None,
    struct = None,
    struct_proxy = None,
    evaluated = None,
    default_val = None,
    default_eval = None,
    init_val = None,
    preset_vals = None,
    doc = None,
    loc = None ):

    if 'schema' in namespace:
      # attempt to take general schema arguments from class body namespace
      schema = namespace.pop('schema')

      if not is_mapping( schema ):
        raise SchemaDefinitionError(
          f"`schema` in schema class body must be a mapping: {schema}",
          loc = loc )

      kwargs = dict(
        tag_key = tag_key,
        tag = tag,
        declared = declared,
        struct = struct,
        struct_proxy = struct_proxy,
        evaluated = evaluated,
        default_val = default_val,
        default_eval = default_eval,
        init_val = init_val,
        preset_vals = preset_vals,
        doc = doc,
        loc = loc )

      for k, v in kwargs.items():

        if k in schema:
          if v is not None and v != schema[k]:
            raise SchemaDefinitionError(
              f"schema '{k}' = {v} redefined in class body: {schema[k]}",
              loc = loc )

          kwargs[k] = schema[k]

      tag_key = kwargs['tag_key']
      tag = kwargs['tag']
      declared = kwargs['declared']
      struct = kwargs['struct']
      struct_proxy = kwargs['struct_proxy']
      evaluated = kwargs['evaluated']
      default_val = kwargs['default_val']
      default_eval = kwargs['default_eval']
      init_val = kwargs['init_val']
      preset_vals = kwargs['preset_vals']
      doc = kwargs['doc']
      loc = kwargs['loc']

    if declared is not None:
      # defines the given schema declared

      if not isinstance( declared, SchemaStructDeclared ):
        raise SchemaDefinitionError(
          f"`declared` must be subclass of SchemaStructDeclared: {fmt_base_or_type(declared)}",
          loc = loc )

      if tag_key is not None and tag_key != declared.tag_key:
        raise SchemaDefinitionError(
          f"declared `tag_key` {declared.tag_key} redefined: {tag_key}",
          loc = loc )

      if tag is not None and tag != declared.tag:
        raise SchemaDefinitionError(
          f"declared `tag` {declared.tag} redefined: {tag}",
          loc = loc )

      tag_key = declared.tag_key
      tag = declared.tag


    if struct is None:
      struct = dict()

    if not isinstance( struct, dict ):
      if is_sequence( struct ) or is_mapping( struct ):
        try:
          struct = dict( struct )

        except BaseException as e:
          raise SchemaDefinitionError(
            f"Schema struct must be an ordered dictionary, or items list: {struct}",
            loc = loc,
            hints = e )

      else:
        raise SchemaDefinitionError(
          f"Schema struct must be a mapping or sequence of key-value pairs: {struct}",
          loc = loc )

    else:
      struct = copy(struct)

    # attempt to extract any struct schema items from class body namespace
    ns_struct_keys = list()

    for k, v in namespace.items():
      if is_schema( v ):
        struct[k] = v

        ns_struct_keys.append(k)

    for k in ns_struct_keys:
      # remove extracted items from namespace
      namespace.pop(k)

    # check base classes for schema definitions
    base_schema = None

    for base in bases:

      if isinstance( base, SchemaStructProxy ):
        if base_schema:
          raise SchemaDefinitionError(
            f"Schema multiple inheritance is not supported: {bases}",
            loc = loc )

        base_schema = base

    schema_bases = tuple()

    if base_schema:
      _base_schema = base_schema.schema

      schema_bases = ( _base_schema, )
      # use values from base schema to fill in those not specified

      # starts with all struct items of base schema
      _struct = dict(_base_schema.struct)

      # add/overwrite any items given for derived schema
      # NOTE: intended to preserve the original order of keys from the base schema
      # for any keys the derived class overwites

      for k,v in struct.items():
        _struct[k] = v

      struct = _struct

      if tag_key is None:
        # attempt to use base class tag_key
        tag_key = base_schema.tag_key

      if tag is None:
        # attempt to use base class tag
        tag = base_schema.tag

      if struct_proxy is None:
        struct_proxy = _base_schema._p_struct_proxy

      if default_val is None:
        default_val = _base_schema._p_default_val

      if default_eval is None:
        default_eval = _base_schema._p_default_eval

      if init_val is None:
        init_val = _base_schema._p_init_val

      # TODO: this would currently make the docs inherit unrelated docstrings
      # if doc is None and not namespace.get('__doc__', None):
      #   doc = _base_schema._p_doc

    schema_namespace = dict()

    if '__doc__' in namespace:
      schema_namespace['__doc__'] = namespace.pop('__doc__')

    schema_namespace['__module__'] = namespace['__module__']
    schema_name = name + '_schema'

    schema = SchemaStruct(
      schema_name,
      schema_bases,
      schema_namespace,
      tag_key = tag_key,
      tag = tag,
      struct = struct,
      struct_proxy = struct_proxy,
      declared = declared,
      evaluated = evaluated,
      default_val = default_val,
      default_eval = default_eval,
      init_val = init_val,
      preset_vals = preset_vals,
      doc = doc,
      loc = loc )

    module = importlib.import_module( namespace['__module__'] )

    if hasattr( module, schema_name ):
      warnings.warn(f"Overriding module attribute with class definition: {module}, {schema_name}")

    setattr( module, schema_name, schema )

    props = dict()

    #...........................................................................
    for k, v in schema.struct.items():
      props[k] = SchemaProperty(
        schema = v,
        name = k )

    props[ schema.tag_key ] = ConstProperty(
      name = schema.tag_key )


    namespace = {
      **namespace,
      **props,
      '_p_tag_key' : schema.tag_key,
      '_p_tag' : schema.tag }

    cls = super().__new__(
      mcls,
      name,
      bases,
      namespace,
      schema = schema )

    return cls

  #-----------------------------------------------------------------------------
  def __init__( cls,
    name,
    bases,
    namespace,
    tag_key = None,
    tag = None,
    declared = None,
    struct = None,
    struct_proxy = None,
    evaluated = None,
    default_val = None,
    default_eval = None,
    init_val = None,
    doc = None,
    loc = None ):

    # NOTE: schema argument not used by __init__
    super().__init__( name, bases, namespace, schema = None )

    # set this class as the valued_type now that it is fully initialized
    cls.schema.valued_type = cls

  #-----------------------------------------------------------------------------
  @property
  def tag_key( cls ):
    """str : Key used to obtain the tag for this struct type
    """
    return cls._p_tag_key

  #-----------------------------------------------------------------------------
  @property
  def tag( cls ):
    """str : Tag for this struct type
    """
    return cls._p_tag

  #-----------------------------------------------------------------------------
  # NOTE: already implemented by StructValued __init__?
  # def __call__( cls,
  #   *args,
  #   **kwargs ):
  #
  #   if len(args) == 0:
  #     val = None
  #     loc = None
  #   else:
  #     if len(args) == 1:
  #       val = args[0]
  #       loc = None
  #     elif len(args) == 2:
  #       val, loc = args
  #     else:
  #       raise ValueError(
  #         f"positional arguments must be at most `(val, loc)`."
  #         " All keyword arguments are interpreted as items of `val`")
  #
  #   if kwargs:
  #     if val:
  #       val = {**val, **kwargs}
  #     else:
  #       val = kwargs
  #
  #   return cls.schema.decode(
  #     val = val,
  #     loc = loc )



import pytest
from pytest import raises

from pprint import pprint

import itertools

from partis.schema import (
  required,
  optional,
  BoolPrimDeclared,
  BoolPrim,
  IntPrimDeclared,
  IntPrim,
  FloatPrimDeclared,
  FloatPrim,
  StrPrimDeclared,
  StrPrim,
  SeqPrimDeclared,
  SeqPrim,
  MapPrimDeclared,
  MapPrim,
  UnionPrimDeclared,
  UnionPrim,
  StructValued,
  PJCEvaluated,
  SchemaError,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_struct_valued,
  any_schema,
  SchemaDefinitionError,
  SchemaDeclaredError,
  schema_declared )

from partis.schema.serialize.yaml import (
  loads,
  dumps )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_prim_declared():
  scalar_cases = [
    ( IntPrimDeclared, IntPrim ),
    ( BoolPrimDeclared, BoolPrim ),
    ( FloatPrimDeclared, FloatPrim ),
    ( StrPrimDeclared, StrPrim ) ]

  item_cases = [
    ( SeqPrimDeclared, SeqPrim ),
    ( MapPrimDeclared, MapPrim ) ]

  for decl, prim in scalar_cases:
    d = decl()

    with raises( SchemaDeclaredError ):
      d.schema

    s = prim( declared = d )

    assert d.schema is s

  for decl, prim in item_cases:
    for idecl, iprim in scalar_cases:

      d = decl()
      di = idecl()

      with raises( SchemaDeclaredError ):
        d.schema

      s = prim(
        declared = d,
        item = di )

      assert not d.schema_defined
      assert not s.schema_defined


      si = iprim( declared = di )

      assert d.schema is s

      assert s.item.schema is si

      assert d.schema.item.schema is si

    d = UnionPrimDeclared()

    with raises( SchemaDeclaredError ):
      d.schema

    s = UnionPrim(
      declared = d,
      cases = [
        BoolPrim(),
        IntPrim() ]  )

    assert d.schema is s

    cases = [ a[0]() for a in scalar_cases[1:] ]

    s = UnionPrim(
      cases = cases )

    assert not s.schema_defined

    _cases = [
      a[1]( declared = d ) for d, a in zip( cases, scalar_cases[1:] ) ]

    assert s.schema_defined

    for d, c in zip( cases, _cases ):
      assert d.schema is c

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_struct_declared():

  my_declared = schema_declared( tag = 'my_type', tag_key = 'my_tag' )

  with raises( SchemaDeclaredError ):
    # schema not yet defined
    s = my_declared.schema

  with raises( SchemaError ):
    # cannot instantiate a schema declaration
    v = my_declared( val = dict() )

  with raises( SchemaDefinitionError ):

    # test using metaclass arguments
    class MyType1( StructValued,
      tag = 'not_my_type',
      declared = my_declared ):
      pass

  class MyType2( StructValued,
    declared = my_declared ):
    pass

  assert my_declared.schema is MyType2.schema
  assert MyType2.tag_key == 'my_tag'
  assert MyType2.tag == 'my_type'



# -*- coding: UTF-8 -*-

import logging
log = logging.getLogger(__name__)

from copy import copy
import inspect
import re

from partis.utils import (
  odict,
  cached_property,
  fmt_base_or_type,
  fmt_iterable_or,
  fmt_attr_doc,
  _fmt_class_name,
  fmt_class_name,
  indent_lines,
  isinstance_any )

from partis.pyproj import (
  hash_sha256 )

from partis.utils.special import (
  required,
  optional,
  derived )

from partis.schema_meta.base import (
  SchemaError,
  SchemaNameError,
  SchemaDeclaredError,
  SchemaDefinitionError,
  SchemaValidationError,
  SchemaHint,
  Bias,
  Loc,
  assert_valid_name,
  is_bool,
  is_numeric,
  is_string,
  is_sequence,
  is_mapping,
  is_schema_prim,
  is_schema_struct,
  is_schema,
  is_schema_declared,
  is_schema_struct_valued,
  is_similar_value_type,
  is_evaluated,
  is_valued,
  is_optional,
  is_required )

from partis.schema_meta.schema import (
  fmt_schema_typename )

from partis.schema.valued import (
  Valued )

from partis.schema.eval import (
  NotEvaluated )

from . import (
  SchemaPrimDeclared,
  SchemaPrim,
  PassPrim,
  BoolPrimDeclared,
  BoolPrim,
  IntPrimDeclared,
  IntPrim,
  FloatPrimDeclared,
  FloatPrim,
  StrPrimDeclared,
  StrPrim,
  SeqPrimDeclared,
  SeqPrim,
  MapPrimDeclared,
  MapPrim )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class UnionPrimDeclared( SchemaPrimDeclared ):
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def validate_union_cases( cases, loc ):

  _bool_prim = None
  _numeric_prim = None
  _string_prim = None
  _list_prim = None
  _dict_prim = None
  _schemas = list()
  _evaluated_cases = list()
  _schema_deps = list()

  proxy_cases = list()
  proxies = list()

  for case in cases:
    if case.schema_defined and is_schema_struct( case ) and case.schema.struct_proxy:
      proxy = case.schema.struct[case.schema.struct_proxy]

      if is_schema_struct( proxy ) and proxy.schema.struct_proxy:
        raise SchemaDefinitionError(
          f"`UnionPrim` may not have `Schema` case with a nested `struct_proxy`",
          loc = loc,
          hints = str(case) )

      proxy_cases.append(case)
      proxies.append(proxy)

  for case in cases + proxies:
    if not case.schema_defined:
      _schema_deps.append(case)

    # these conditions are to prevent ambiguities when blind coding 'simple'
    # data that may be interpreted in more than one way
    if is_schema_prim( case ):

      if case.schema_defined:
        _evaluated_cases.append( case.schema.evaluated )

      if isinstance_any( case, [ BoolPrim, BoolPrimDeclared ] ):
        if _bool_prim is not None:
          raise SchemaDefinitionError(
            f"`UnionPrim` may only have one `BoolPrim` case",
            loc = loc,
            hints = [ SchemaHint(f"case: {i}", hints = str(c) ) for i,c in enumerate(cases) ])

        _bool_prim = case

      elif isinstance_any( case, [ IntPrim, IntPrimDeclared, FloatPrim, FloatPrimDeclared ] ):
        if _numeric_prim is not None:
          raise SchemaDefinitionError(
            f"`UnionPrim` may only have one numeric {{`IntPrim`, `FloatPrim`}} case",
            loc = loc,
            hints = [ SchemaHint(f"case: {i}", hints = str(c) ) for i,c in enumerate(cases) ])

        _numeric_prim = case

      elif isinstance_any( case, [ StrPrim, StrPrimDeclared ] ):
        if _string_prim is not None:
          raise SchemaDefinitionError(
            f"`UnionPrim` may only have one `StrPrim` case",
            loc = loc,
            hints = [ SchemaHint(f"case: {i}", hints = str(c) ) for i,c in enumerate(cases) ])

        _string_prim = case

      elif isinstance_any( case, [ SeqPrim, SeqPrimDeclared ] ):
        if _list_prim is not None:
          raise SchemaDefinitionError(
            f"`UnionPrim` may only have one `SeqPrim` case",
            loc = loc,
            hints = [ SchemaHint(f"case: {i}", hints = str(c) ) for i,c in enumerate(cases) ] )

        _list_prim = case

      elif isinstance_any( case, [ MapPrim, MapPrimDeclared ] ):
        if len(_schemas) > 0:
          raise SchemaDefinitionError(
            f"`UnionPrim` may not have both `SchemaStruct` and `MapPrim` cases",
            loc = loc,
            hints = [ SchemaHint(f"case: {i}", hints = str(c) ) for i,c in enumerate(cases) ] )

        elif _dict_prim is not None:
          raise SchemaDefinitionError(
            f"`UnionPrim` may only have one `MapPrim` case",
            hints = [ SchemaHint(f"case: {i}", hints = str(c) ) for i,c in enumerate(cases) ] )

        _dict_prim = case

      elif isinstance_any( case, [ UnionPrim, UnionPrimDeclared ] ):
        raise SchemaDefinitionError(
          f"`UnionPrim` may not contain any `UnionPrim` cases",
          loc = loc,
          hints = [ SchemaHint(f"case: {i}", hints = str(c) ) for i,c in enumerate(cases) ] )

      elif isinstance( case, PassPrim ):
        raise SchemaDefinitionError(
          f"`UnionPrim` may not contain any `PassPrim` cases",
          loc = loc,
          hints = [ SchemaHint(f"case: {i}", hints = str(c) ) for i,c in enumerate(cases) ] )

      else:
        assert False

    elif is_schema_struct( case ):

      if case.schema_defined:

        _evaluated_cases.append( case.schema.evaluated )

        # if case.schema.struct_proxy:
        #   raise SchemaDefinitionError(
        #     f"`UnionPrim` may not have `Schema` case with a `struct_proxy`",
        #     loc = loc,
        #     hints = str(case) )

        if _dict_prim is not None:
          raise SchemaDefinitionError(
            f"`UnionPrim` may not have both `SchemaStruct` and `MapPrim` cases",
            loc = loc,
            hints = [ SchemaHint(f"case: {i}", hints = str(c) ) for i,c in enumerate(cases) ] )

        if len(_schemas) > 1 and _schemas[0].tag_key != case.tag_key:
          raise SchemaDefinitionError(
            f"`UnionPrim` may not have Schemas cases with multiple values for `tag_key`",
            loc = loc,
            hints = [ SchemaHint(f"case: {i}", hints = str(c) ) for i,c in enumerate(cases) ] )

      # multiple schemas may be in the union, since they can be differentiated
      # by the `tag_key`
      _schemas.append( case )

    else:
      assert False

  return ( _bool_prim,
    _numeric_prim,
    _string_prim,
    _list_prim,
    _dict_prim,
    _schemas,
    proxy_cases,
    proxies,
    _evaluated_cases,
    _schema_deps )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class UnionPrim( SchemaPrim ):
  """Primitive for union of types values

  Parameters
  ----------
  cases : list[ :class:`SchemaStruct <partis.schema.struct.SchemaStruct>` | :class:`SchemaPrim <partis.schema.prim.base.SchemaPrim>` ]
    Schemas that are considered for decoding a value.
    The union of schemas may not be ambiguous, restricting the combinations to

      - Max of one case of :class:`BoolPrim <partis.schema.prim.bool_prim.BoolPrim>`.

      - Max of one numeric case of either :class:`IntPrim <partis.schema.prim.int_prim.IntPrim>`
        or :class:`FloatPrim <partis.schema.prim.float_prim.FloatPrim>`.

      - Max of one case of :class:`StrPrim <partis.schema.prim.str_prim.StrPrim>`.

      - Max of one case of :class:`SeqPrim <partis.schema.prim.seq_prim.SeqPrim>`.

      - Max of one case of :class:`MapPrim <partis.schema.prim.map_prim.MapPrim>`.

      - May not have cases of both :class:`MapPrim <partis.schema.prim.map_prim.MapPrim>`
        and :class:`SchemaStruct <partis.schema.struct.SchemaStruct>`.

      - All cases of :class:`SchemaStruct <partis.schema.struct.SchemaStruct>` must be defined
        with the same `tag_key`, and no cases with a nested `struct_proxy`.

      - May not have any case of :class:`PassPrim <partis.schema.prim.pass_prim.PassPrim>`

  default_case : NoneType | int
    Case to use for default value when source data is not available, including
    if the case's `default_val` is required or optional.

  **kwargs : arguments passed to :class:`SchemaPrim <partis.schema.prim.base.SchemaPrim>`
  """

  #-----------------------------------------------------------------------------
  def __new__( mcls,
    name,
    bases,
    namespace,
    cases,
    default_case = None,
    **kwargs ):

    loc = kwargs.get('loc', None)

    if not is_sequence( cases ):
      raise SchemaDefinitionError(
        f"`UnionPrim` cases must be list of `SchemaPrim` or `Schema`: {type(cases).__name__}",
        loc = loc )

    cases = list( cases )

    for i, case in enumerate(cases):
      if not is_schema( case ):
        raise SchemaDefinitionError(
          f"`UnionPrim` case '{i}' must be instance of `SchemaPrim` or `Schema`: {case}",
          loc = loc )

    if len(cases) < 1:
      raise SchemaDefinitionError(
        f"`UnionPrim` must have at least one case",
        loc = loc )

    if default_case is None:
      default_case = required

    if not is_required( default_case ):
      if not is_numeric( default_case ):
        raise SchemaDefinitionError(
          f"`default_case` must be an integer: {default_case}",
          loc = loc )

      # TODO: check that default_val is 'required', since it would be ambiguous
      # to specify both default_val and default_case

      default_case = int(default_case)

      if default_case < 0 or default_case >= len(cases):
        raise SchemaDefinitionError(
          f"`default_case` invalid with {len(cases)} cases: {default_case}",
          loc = loc )

    ( _bool_prim,
      _numeric_prim,
      _string_prim,
      _list_prim,
      _dict_prim,
      _schemas,
      proxy_cases,
      proxies,
      _evaluated_cases,
      _schema_deps ) = validate_union_cases(
        cases = cases,
        loc = loc )

    #...........................................................................
    _schema_tags = [ c.tag for c in _schemas ]

    kwargs['schema_deps'] = kwargs.get('schema_deps', list()) + _schema_deps

    namespace = { **namespace, **dict(
      _p_bool_prim = _bool_prim,
      _p_numeric_prim = _numeric_prim,
      _p_string_prim = _string_prim,
      _p_list_prim = _list_prim,
      _p_dict_prim = _dict_prim,
      _p_schemas = _schemas,
      _p_proxy_cases = proxy_cases,
      _p_proxies = proxies,
      _p_schema_tags = _schema_tags,
      _p_default_case = default_case,
      _p_cases = cases ) }

    return super().__new__( mcls,
      name = name,
      bases = bases,
      namespace = namespace,
      valued_type = Valued,
      **kwargs )

  #-----------------------------------------------------------------------------
  def __init__( cls,
    name,
    bases,
    namespace,
    cases,
    default_case = None,
    **kwargs ):


    super().__init__(
      name = name,
      bases = bases,
      namespace = namespace,
      **kwargs )

  #-----------------------------------------------------------------------------
  def _get_attr_doc_lines( cls,
    noindex = False,
    depth = 0,
    max_depth = 2 ):

    lines = super()._get_attr_doc_lines(
      noindex = noindex,
      depth = depth,
      max_depth = max_depth )

    lines.extend([
      fmt_attr_doc(
        name = attr,
        typename = type( getattr( cls, attr ) ),
        val = getattr( cls, attr ),
        prefix = 'schema',
        noindex = noindex )
      for attr in [
        'default_case' ] ])

    case_lines = list()

    for i, case in enumerate(cls.cases):
      # schema = case.schema

      case_lines.append( f"{i} {fmt_schema_typename(case, fmt_class_name)}" )

      # case_lines.append( indent_lines( 2, schema._get_doc(
      #   # always specify noindex since the item schema already documented
      #   noindex = True,
      #   depth = depth + 1,
      #   max_depth = max_depth ) ) )

      case_lines.append('')

    lines.append(
      fmt_attr_doc(
        name = 'cases',
        typename = f'list[~partis.schema_meta.schema.Schema]',
        val = ...,
        prefix = 'schema',
        doc = '\n'.join(case_lines),
        noindex = noindex ) )

    return lines

  #-----------------------------------------------------------------------------
  def _schema_hash_children( cls ):
    return list(cls.cases)

  #-----------------------------------------------------------------------------
  @property
  def cases( cls ):
    return cls._p_cases

  #-----------------------------------------------------------------------------
  @property
  def default_case( cls ):
    return cls._p_default_case

  #-----------------------------------------------------------------------------
  @cached_property
  def default_val( cls ):
    val = super().default_val

    if (
      ( is_optional(val) or is_required(val) )
      and not is_required(cls.default_case) ):

      case = cls.cases[ cls.default_case ].schema
      val = case.default_val

      if (
        is_mapping(val)
        and is_schema_struct(case)
        and case.tag_key in val
        and len(cls._p_schemas) == 1 ):
          # tag is not needed when schema is not ambiguous
          # NOTE: this is an exception to the rule that the default_val includes the
          # tag, since there cannot be a union directly in another union.
          val.pop( case.tag_key )

    return val

  #-----------------------------------------------------------------------------
  @cached_property
  def default_eval( cls ):
    val = super().default_eval

    if not is_required( cls.default_case ) and (
      is_optional( val )
      or is_required( val ) ):

      case = cls.cases[ cls.default_case ].schema
      val = case.default_eval

      if (
        is_mapping(val)
        and is_schema_struct(case)
        and case.tag_key in val
        and len(cls._p_schemas) == 1 ):
          # tag is not needed when schema is not ambiguous
          val.pop( case.tag_key )

    return val

  #-----------------------------------------------------------------------------
  @cached_property
  def init_val( cls ):
    val = super().init_val

    if val is not None:
      return val

    init_case = cls.default_case

    if is_required( init_case ):
      init_case = 0

    case = cls.cases[ init_case ].schema
    val = case.init_val

    if is_schema_struct(case) and len(cls._p_schemas) == 1:
      val.pop( case.tag_key )

    return val

  #-----------------------------------------------------------------------------
  @property
  def hints( cls ):
    hints = super().hints

    hints.append(f"default_case: {cls.default_case}")

    _hints = list()

    for i, c in enumerate(cls.cases):
      _hints.append( SchemaHint( f"{i}: {c.schema.__module__}.{c.schema.__name__}" ) )

      # if is_schema_declared(c):
      #   _hints.append( SchemaHint( f"{i}: {case.schema.__module__}.{case.schema.__name__}" ) )
      #
      # else:
      #   _hints.append( SchemaHint( f"{i}:", hints = c.schema.hints ) )

    hints.append( SchemaHint(
      f"cases: list[schema]",
      hints = _hints ) )

    return hints

  #-----------------------------------------------------------------------------
  def _schema_resolved( cls ):

    ( _bool_prim,
      _numeric_prim,
      _string_prim,
      _list_prim,
      _dict_prim,
      _schemas,
      proxy_cases,
      proxies,
      _evaluated_cases,
      _schema_deps ) = validate_union_cases(
        cases = cls._p_cases,
        loc = cls.loc )

    cls._p_bool_prim = _bool_prim
    cls._p_numeric_prim = _numeric_prim
    cls._p_string_prim = _string_prim
    cls._p_list_prim = _list_prim
    cls._p_dict_prim = _dict_prim
    cls._schemas = _schemas
    cls._p_proxy_cases = proxy_cases
    cls._p_proxies = proxies

    _evaluated_cases.append( cls.evaluated )

    eval_options = list()

    for eval in _evaluated_cases:
      if ( eval is not NotEvaluated ) and ( eval not in eval_options ):
        eval_options.append(eval)

    if len(eval_options) == 0:
      evaluated = NotEvaluated

    elif len(eval_options) == 1:
      evaluated = eval_options[0]

    else:
      evaluated = eval_options[0].subclass(
        name = cls.__name__ + '_evaluated',
        module = cls.__module__,
        bases = eval_options[1:] )

    cls._p_evaluated = evaluated

    super()._schema_resolved()

  #-----------------------------------------------------------------------------
  def encode( cls,
    val = None,
    loc = None,
    no_defaults = None ):

    val, loc, validate = cls._encode(
      val = val,
      loc = loc,
      no_defaults = no_defaults )

    if not validate:
      return val

    val = cls.validate(
      val = val,
      loc = loc )[0]

    schema = None
    tag_key = None
    tag_key_optional = False

    if is_bool( val ) and cls._p_bool_prim is not None:
      schema = cls._p_bool_prim.schema

    elif is_numeric( val ) and cls._p_numeric_prim is not None:
      schema = cls._p_numeric_prim.schema

    elif is_string( val ) and cls._p_string_prim is not None:
      schema = cls._p_string_prim.schema

    elif is_sequence( val ) and cls._p_list_prim is not None:
      schema = cls._p_list_prim.schema

    elif is_mapping( val ) and cls._p_dict_prim is not None:
      schema = cls._p_dict_prim.schema

    elif is_mapping( val ) and len(cls._p_schemas) > 0:
      tag_key = cls._p_schemas[0].tag_key

      if len(cls._p_schemas) == 1:
        # only one schema to try
        schema = cls._p_schemas[0].schema
        tag_key_optional = True

      else:
        # requires tag to resolve ambiguity

        if tag_key not in val:
          # if the schema is ambiguous, must have given the tag value
          _tags = fmt_iterable_or( cls._p_schema_tags )

          raise SchemaValidationError(
            f"Union of schemas item '{tag_key}' must be any of {_tags}",
            loc = loc )

        tag = val[tag_key]

        if tag not in cls._p_schema_tags:
          _tags = fmt_iterable_or( cls._p_schema_tags )

          raise SchemaValidationError(
            f"Union of schemas item '{tag_key}' must be any of {_tags}: {tag}",
            loc = loc )

        schema = cls._p_schemas[ cls._p_schema_tags.index( tag ) ].schema

    else:
      raise SchemaValidationError(
        f"`{cls.__name__}` union case not found for value: {fmt_base_or_type(val)}",
        loc = None,
        hints = [ SchemaHint(
          f"case {i}: `{c.schema.__name__}` <{type(c.schema).__name__}>" )
          for i,c in enumerate(cls.cases) ] )

    val = schema.encode(
      val = val,
      no_defaults = no_defaults )

    # if no_defaults:
    #   default_val = cls.default_val
    #
    #   if not ( is_optional(default_val) or is_required(default_val) ):
    #
    #     if is_similar_value_type(val, default_val) and val == default_val:
    #       # NOTE: is_similar_value_type is needed to prevent removal of values
    #       # that are "equal", but have a different type.
    #       # E.G: 0.0 == False, but float != bool.
    #       val = None
    #
    #   if val is None and (
    #     is_required(default_val)
    #     or is_optional(default_val)
    #     or is_required(cls.default_case)
    #     or schema is not cls.cases[ cls.default_case ].schema ):
    #
    #     val = schema.default_val

    if is_mapping(val) and tag_key in val and tag_key_optional:
      val.pop( tag_key )

    return val

  #-----------------------------------------------------------------------------
  def decode( cls,
    val = None,
    loc = None,
    bias = None ):

    bias = Bias(bias)

    val, loc, validate, bias = cls._decode(
      val = val,
      loc = loc,
      bias = bias )

    if not validate:
      return val

    val, bias = cls.validate(
      val = val,
      loc = loc,
      bias = bias )

    if is_bool( val ) and cls._p_bool_prim is not None:
      if cls._p_bool_prim in cls._p_proxies:
        return cls._p_proxy_cases[cls._p_proxies.index(cls._p_bool_prim)].schema.decode(
          val = val,
          loc = loc,
          bias = bias )
      else:
        return cls._p_bool_prim.schema.decode(
          val = val,
          loc = loc,
          bias = bias  )

    if is_numeric( val ) and cls._p_numeric_prim is not None:
      if cls._p_numeric_prim in cls._p_proxies:
        return cls._p_proxy_cases[cls._p_proxies.index(cls._p_numeric_prim)].schema.decode(
          val = val,
          loc = loc,
          bias = bias  )
      else:
        return cls._p_numeric_prim.schema.decode(
          val = val,
          loc = loc,
          bias = bias )

    if is_string( val ) and cls._p_string_prim is not None:
      if cls._p_string_prim in cls._p_proxies:
        return cls._p_proxy_cases[cls._p_proxies.index(cls._p_string_prim)].schema.decode(
          val = val,
          loc = loc,
          bias = bias )
      else:
        return cls._p_string_prim.schema.decode(
          val = val,
          loc = loc,
          bias = bias )

    if is_sequence( val ) and cls._p_list_prim is not None:
      if cls._p_list_prim in cls._p_proxies:
        return cls._p_proxy_cases[cls._p_proxies.index(cls._p_list_prim)].schema.decode(
          val = val,
          loc = loc,
          bias = bias )
      else:
        return cls._p_list_prim.schema.decode(
          val = val,
          loc = loc,
          bias = bias )

    if is_mapping( val ) and cls._p_dict_prim is not None:
      if cls._p_dict_prim is not None:
        return cls._p_dict_prim.schema.decode(
          val = val,
          loc = loc,
          bias = bias )

    if is_mapping( val ) and len(cls._p_schemas) > 0:

      if len(cls._p_schemas) == 1:
        # only one schema to try
        return cls._p_schemas[0].schema.decode(
          val = val,
          loc = loc,
          bias = bias )

      else:
        # requires tag to resolve ambiguity

        tag_key = cls._p_schemas[0].tag_key

        if tag_key not in val:
          # if the schema is ambiguous, must have given the tag value
          _tags = fmt_iterable_or( cls._p_schema_tags )

          raise SchemaValidationError(
            f"Union of schemas item '{tag_key}' must be any of {_tags}",
            loc = loc )


        tag = val[tag_key]

        if tag not in cls._p_schema_tags:
          _tags = fmt_iterable_or( cls._p_schema_tags )

          raise SchemaValidationError(
            f"Union of schemas item '{tag_key}' must be any of {_tags}: {tag}",
            loc = loc )

        schema = cls._p_schemas[ cls._p_schema_tags.index( tag ) ].schema

        return schema.decode(
          val = val,
          loc = loc,
          bias = bias )

    raise SchemaValidationError(
      f"`{cls.__name__}` union case not found for value: {fmt_base_or_type(val)}",
      loc = loc,
      hints = [ SchemaHint(
        f"case {i}: `{c.schema.__name__}` <{type(c.schema).__name__}>" )
        for i,c in enumerate(cls.cases) ] )
